#!/usr/bin/perl -w

use strict;
#use Net::LDAP;
use HTML::Strip;
use CGI;
use constant REVISION => q$Revision: 1.31 $;

# Set version string
my $VERSION = "1.0-pre10 " . REVISION;

# LDAP info
my $LDAP_SERVER = "ldap.domain.com";
my $LDAP_PORT = 389;
my $LDAP_BINDDN = "cn=Manager,dc=domain,dc=com";
my $LDAP_BINDPW = "mypassword";
my $LDAP_VIRTDN = "ou=VirtUserTable,ou=Mail,dc=domain,dc=com";

# Script location
my $SCRIPT = "/cgi-bin/spasm.cgi";

# Log location (optional, must be writable by webserver process)
#my $USERLOG = "/var/log/httpd/spasm.log";

# Image URL
my $GFX_URL = "/gfx";

# Directories
my $SPASM_QDIR = "/etc/mail/spasm/quarantine";
my $SPASM_LDIR = "/etc/mail/spasm/logs";
my $SPASM_PREFS = "/etc/mail/spasm/prefs";

# Local Blacklist
my $LOCALBL = "localbl";

# Global IP Blacklist
my $GLOBAL_IP_BL = "global_ip_bl";

# Location of spasm_cgi_app and optional authuser programs
my $SPASM_CGI_APP = "/var/www/bin/spasm/spasm_cgi_app";
my $SPASM_AUTH_APP = "/var/www/bin/spasm/authuser";

# Default Spasm domain
my $DEFAULT_DOMAIN = "theasylum.org";

# Location of non-spasm programs
my $SENDMAIL_BIN = "/usr/sbin/sendmail";
my $MAIL_BIN = "/bin/mail";

# Date Formats
my @DATE_FORMATS = (
"Day MM DD hh:mm:ss YYYY",
"MM DD hh:mm:ss YYYY",
"MM DD hh:mm:ss",
"MM DD"
);

# Cookies
my $COOKIE_ID = "";
my $COOKIE_PASS = "";
my $COOKIE_ISP = "";

# ISPs: "value" => "display name"
my %ISPS = ();
 # Example multi-ISP setup:
# my %ISPS = (
# "domain1" => { domain => "domain1.com", order => 0 },
# "domain2" => { domain => "domain2.com", order => 1 },
# "domain3" => { domain => "domain3.com", order => 2 }
# );

# CSS
my $CSS = <<EOF;
      BODY { background: white; color: black; font-family: "times new roman", serif; font-size: 12pt; }
      H1 { font-size: 13pt; text-align: center; }
      H2 { font-size: 10pt; text-align: center; }
      IMG { border: none; }
      TABLE { margin-left: auto; margin-right: auto; }
      FORM { margin-bottom: 0; }
      A:link { text-decoration: none; color: #004080; }
      A:visited { text-decoration: none; color: #004080; }
      A:hover { text-decoration: none; color: #990000; background: #ddddff; }
      A:active { text-decoration: none; color: #004080; }
EOF
  
# Spasm statuses
my @STATUS = (
  "Off",	# SPASM_OFF = 0
  "Tagged",	# SPASM_TAG = 1
  "Quarantined",# SPASM_QUARANTINE = 2
  "",
  "Rejected"	# SPASM_REJECT = 4
);

# Spasm filters, filter groups, and filter group settings
my @SF = (
  "SF_WHITELIST_ONLY",
  "SF_VIRUS_SCAN",
  "SF_PERSONAL_BLACKLIST",
  "SF_SPAMTRAP",
  "SF_LOCALFILTER",
  "SF_RDNS",
  "SF_RDNS_APNIC",
  "SF_RDNS_RIPE",
  "SF_RDNS_LACNIC",
  "SF_HELO",
  "SF_MAIL_HOST",
  "SF_EXE_ATTACH",
  "SF_SPAMASSASSIN",
  "SA_REQUIRED_HITS",
  "SF_SPAMCOP",
  "SF_SBL",
  "SF_SORBS",
  "SF_SPEWS",
  "SF_POSTMASTERRFC",
  "SF_DSNRFC",
  "SF_ABUSERFC",
  "SF_WHOISRFC"
);

my @FILTERGROUP = (
  "Off",
  "Low",
  "Medium",
  "High",
  "All",
  "Custom"
);

my %FG_LOW = (
  "SF_WHITELIST_ONLY", 0,
  "SF_VIRUS_SCAN", 10,
  "SF_PERSONAL_BLACKLIST", 10,
  "SF_SPAMTRAP", 0,
  "SF_LOCALFILTER", 0,
  "SF_RDNS", 0,
  "SF_RDNS_APNIC", 0,
  "SF_RDNS_RIPE", 0,
  "SF_RDNS_LACNIC", 0,
  "SF_HELO", 0,
  "SF_MAIL_HOST", 0,
  "SF_EXE_ATTACH", 0,
  "SF_SPAMASSASSIN", 0,
  "SA_REQUIRED_HITS", 5,
  "SF_SPAMCOP", 0,
  "SF_SBL", 10,
  "SF_SORBS", 10,
  "SF_SPEWS", 0,
  "SF_POSTMASTERRFC", 0,
  "SF_DSNRFC", 0,
  "SF_ABUSERFC", 0,
  "SF_WHOISRFC", 0
);

my %FG_MED = (
  "SF_WHITELIST_ONLY", 0,
  "SF_VIRUS_SCAN", 10,
  "SF_PERSONAL_BLACKLIST", 10,
  "SF_SPAMTRAP", 10,
  "SF_LOCALFILTER", 10,
  "SF_RDNS", 0,
  "SF_RDNS_APNIC", 0,
  "SF_RDNS_RIPE", 0,
  "SF_RDNS_LACNIC", 0,
  "SF_HELO", 0,
  "SF_MAIL_HOST", 0,
  "SF_EXE_ATTACH", 0,
  "SF_SPAMASSASSIN", 10,
  "SA_REQUIRED_HITS", 5,
  "SF_SPAMCOP", 10,
  "SF_SBL", 10,
  "SF_SORBS", 10,
  "SF_SPEWS", 0,
  "SF_POSTMASTERRFC", 0,
  "SF_DSNRFC", 0,
  "SF_ABUSERFC", 0,
  "SF_WHOISRFC", 0
);

my %FG_HIGH = (
  "SF_WHITELIST_ONLY", 0,
  "SF_VIRUS_SCAN", 10,
  "SF_PERSONAL_BLACKLIST", 10,
  "SF_SPAMTRAP", 10,
  "SF_LOCALFILTER", 10,
  "SF_RDNS", 10,
  "SF_RDNS_APNIC", 0,
  "SF_RDNS_RIPE", 0,
  "SF_RDNS_LACNIC", 0,
  "SF_HELO", 10,
  "SF_MAIL_HOST", 10,
  "SF_EXE_ATTACH", 0,
  "SF_SPAMASSASSIN", 10,
  "SA_REQUIRED_HITS", 4,
  "SF_SPAMCOP", 10,
  "SF_SBL", 10,
  "SF_SORBS", 10,
  "SF_SPEWS", 0,
  "SF_POSTMASTERRFC", 10,
  "SF_DSNRFC", 10,
  "SF_ABUSERFC", 10,
  "SF_WHOISRFC", 10
);

my %FILTER_HELP = (
  "SF_WHITELIST_ONLY", "This will reject all messages. The purpose of this filter is to only allow email addresses or mail servers which have been explicitly listed in the user\\'s whitelist. The user\\'s whitelist is checked prior to any filter checks, so any message caught by the whitelist will automatically bypass all filters.",
  "SF_VIRUS_SCAN", "This runs the message through Spasm\\'s virus scanner.",
  "SF_PERSONAL_BLACKLIST", "This turns on the user\\'s personal blacklist. Setting this filter to \\'Reject\\' bounces mail matching blacklist entries back to the sender.",
  "SF_SPAMTRAP", "This filter checks the message against the system\\'s spamtrap blacklist. Mail servers are automatically added to this blacklist by Spasm if they have attempted to send email to a \\'spamtrap\\' email address: that is, an email address which has been set up by the Systems Administrator specifically to catch spammers. Such an email address is not advertised openly, and should never receive valid email. Any email sent to such an address, therefore, is assumed to be spam.",
  "SF_LOCALFILTER", "This is nearly identical to SF_PERSONAL_BLACKLIST, except it is system-wide and is maintained by the Systems Administrator. This filter also checks a powerful system-wide list which is maintained by the Systems Administrator.",
  "SF_RDNS", "This is a very aggressive filter, and will often catch legitimate email. This filter checks the IP address of the server trying to send a message. If the IP address does not resolve to a hostname (ie, it does not have a proper DNS entry), it is rejected. A large percentage of spammers use IP addresses that have no DNS entry, but likewise, there are many legitimate mail servers whose Systems Administrators have not set up proper DNS, so use this with caution.",
  "SF_RDNS_APNIC", "This is a less aggressive filter, performing the same filtering as the SF_RDNS filter but restricted to the Asia-Pacific networks.",
  "SF_RDNS_RIPE", "This is a less aggressive filter, performing the same filtering as the SF_RDNS filter but restricted to the European networks.",
  "SF_RDNS_LACNIC", "This is a less aggressive filter, performing the same filtering as the SF_RDNS filter but restricted to the Latin-American networks.",
  "SF_HELO", "This is a moderately aggressive filter, and may catch legitimate email. When one mail server connects to another, it should (but is not always required to) send what is called a \\'HELO/EHLO\\' command. This command allows the first mail server to identify itself to the second mail server, by sending its hostname. This filter checks the hostname given during the HELO/EHLO command against the server\\'s actual hostname (by looking up its IP address). If they do not share at least the second-level domain portion, it is rejected (ie, with the hostname \\'my.mail.domain.com\\', it only checks the \\'domain.com\\' portion). This filter will stop a considerable amount of forged spam. If no HELO/EHLO command is given, this filter will also reject the message.",
  "SF_MAIL_HOST", "This is a very aggressive filter, and will often catch legitimate email. This checks the second-level domain portion of the sender\\'s email address against the second-level domain portion of the sending server\\'s hostname.  If they do not match, the message is rejected. This filter will stop messages which have been relayed through other servers. However, this means it will also stop messages that are legitimately being relayed (eg, virtual domain email).",
  "SF_EXE_ATTACH", "This filter scans the message for an executable attachment.",
  "SF_SPAMASSASSIN", "This runs SpamAssassin, a content-based filter, against the message.",
  "SA_REQUIRED_HITS", "When SpamAssassin scans a message, it returns a numeric score based on its own scoring system. The value of SA_REQUIRED_HITS is checked against the score, and if the score is greater than or equal to the value set here, the message is rejected. A good moderate value is 5, but can be reduced, increasing the chance of catching messages with words like \\'viagra.\\'",
  "SF_SPAMCOP", "This checks the sending server\\'s IP address against the SpamCop database. For more information, see http://www.spamcop.net",
  "SF_SBL", "This checks the sending server\\'s IP address against the Spamhaus database. For more information, see http://www.spamhaus.org",
  "SF_SORBS", "This checks the sending server\\'s IP address against the SORBS database. For more information, see http://www.sorbs.net",
  "SF_SPEWS", "This checks the sending server\\'s IP address against the SPEWS level 2 database. For more information, see http://www.spews.org",
  "SF_POSTMASTERRFC", "The RFC filters check the sending server\\'s IP address or hostname against the rfc-ignorant.org databases. For more information on any of these five filters, see http://www.rfc-clueless.org",
  "SF_DSNRFC", "The RFC filters check the sending server\\'s IP address or hostname against the rfc-ignorant.org databases. For more information on any of these five filters, see http://www.rfc-clueless.org",
  "SF_ABUSERFC", "The RFC filters check the sending server\\'s IP address or hostname against the rfc-ignorant.org databases. For more information on any of these five filters, see http://www.rfc-clueless.org",
);

my %SPASM_HELP = (                                                                                                                                                    
  "ALIASED_ACCOUNTS", "Adding an aliased account tells Spasm to use the primary account\\'s settings and quarantine for mail sent to the aliased account.",
  "use_subject_tagging", "Because some mail clients (notably, Outlook Express) do not allow custom filtering, turning this on will tag the subject line of the message with \\'**** X-spasm ****\\' for any tagged messages, instead of just adding header tags. This way, those mail clients can filter based on the subject line.",
  "use_logging", "This globally turns logging on and off. Logging must also be enabled for each individual filter. By default, the Low, Medium, and High settings have logging turned on for each active filter.",
);

my $cgi = new CGI;

# Grab basic variables from form
my $username = $cgi->param('username')?$cgi->param('username'):"";
my $password = $cgi->param('password')?$cgi->param('password'):"";
my $isp = $cgi->param('isp')?$cgi->param('isp'):"";
my $virtdomain = $cgi->param('virtdomain')?$cgi->param('virtdomain'):"";
my $virtemail = $cgi->param('virtemail')?$cgi->param('virtemail'):"";
my $identity = $cgi->param('identity')?$cgi->param('identity'):"";
my $submit = $cgi->param('submit')?$cgi->param('submit'):"";
my $source = $cgi->param('source')?$cgi->param('source'):"";

# Other variables
my $viewbutton;
my $deliverbutton;
my $whiteballbutton;
my $deletebutton;
my $blackballbutton;
my $delaliasbutton;
my @domains;
my %addresses;

# Auth cookies
unless($username) { $username = $cgi->cookie('spasmLoginID')?$cgi->cookie('spasmLoginID'):""; }
unless($password) { $password = $cgi->cookie('spasmLoginPass')?$cgi->cookie('spasmLoginPass'):""; }
unless($isp) { if(%ISPS) { $isp = $cgi->cookie('spasmLoginISP')?$cgi->cookie('spasmLoginISP'):""; } }

# If we're doing documentation stuff, do it prior to authentication
if($submit eq "viewlocalbl") { viewLocalBL(); exit; }
if($submit eq "searchGlobalBL" or $submit eq "Search Global IP Blacklist") { searchGlobalBL(); exit; }
if($submit eq "deleteGlobalBL") { deleteGlobalBL(); exit; }

# Force lowercase usernames, and strip any leading/trailing spaces
$username = lc($username);
$username =~ s/\s//g;

# Check for any image buttons
my @paramlist = $cgi->param();
for (@paramlist) {
    if(/^viewbutton-(.+?)(?:\.[xy])*$/) { $viewbutton = $1; last; }
    if(/^deliverbutton-(.+?)(?:\.[xy])*$/) { $deliverbutton = $1; last; }
    if(/^whiteballbutton-(.+?)(?:\.[xy])*$/) { $whiteballbutton = $1; last; }
    if(/^deletebutton-(.+?)(?:\.[xy])*$/) { $deletebutton = $1; last; }
    if(/^blackballbutton-(.+?)(?:\.[xy])*$/) { $blackballbutton = $1; last; }
    if(/^delalias-(.+?)(?:\.[xy])*$/) { $delaliasbutton = $1; last; }
}

# Verify that we have required authentication variables:
# username, password, and [optionally] isp
if(%ISPS) {
    unless($username and $password and $isp) { printLoginScreen(); }
} else {
    unless($username and $password) { printLoginScreen(); }
}

# Authenticate user
verifyLogin();

# Read options
readOptions();

# If a generic "Submit" name was used, change it to the name of the actual
# button
if($submit eq "Submit") { $submit = $cgi->param('selectMenu'); }
if(defined($cgi->param('submit2')) and ($cgi->param('submit2') eq "Submit")) { $submit = $cgi->param('selectMenu2'); }

# If this is the initial login, go to the quarantine page
if($submit eq "Login") { viewQuarantine(); }

# Mail Filters Menu
elsif($submit eq "Mail Filters" or $submit eq "Filters") { viewMailFilters(); }
elsif($submit eq "Update" and $source eq "viewMailFilters") { updateMailFilters(); }

# Log Menu
elsif($submit eq "View Selected Entries" or $submit eq "View All Entries") { viewLogEntry(); }
elsif($submit eq "Whitelist Selected Senders") { whitelistEntry(); }
elsif($submit eq "Blacklist Selected Senders") { blacklistEntry(); }
elsif($submit eq "Delete Selected Entries") { deleteLogEntry(); }
elsif($submit eq "Delete All Entries") { deleteLogEntry(); }
elsif($submit eq "Dump Log" or $submit eq "Download Log") { dumpLog(); }
elsif($submit eq "Log" or 
      ($source eq "viewLog" and
       ($submit eq "Next" or $submit eq "Previous" or
	$submit eq "Date" or $submit eq "Sender" or
	$submit eq "Filter" or $submit eq "Status" or $submit eq "Recipient"))) { viewLog(); }

# Quarantine Menu
elsif($viewbutton or $submit eq "View Full Headers" or $submit eq "View Standard Headers") { viewQuarantineFile(); }
elsif($deliverbutton or $submit eq "Send Message" or $submit eq "Send Selected Messages") { deliverQuarantineFile(); }
elsif($whiteballbutton or $submit eq "Send+Whitelist Selected Messages") { whiteballQuarantineFile(); }
elsif($deletebutton or $submit eq "Delete Message" or $submit eq "Delete Selected Messages") { deleteQuarantineFile(); }
elsif($blackballbutton or $submit eq "Delete+Blacklist Selected Messages") { blackballQuarantineFile(); }
elsif($submit eq "Blacklist Sender" or $submit eq "Blacklist Domain" or $submit eq "Blacklist Server") { blacklistEntry(); }
elsif($submit eq "Whitelist Sender" or $submit eq "Whitelist Domain" or $submit eq "Whitelist Server") { whitelistEntry(); }
elsif($submit eq "Delete All Messages") { clearQuarantine(); }
elsif($submit eq "Quarantine" or 
      ($source eq "viewQuarantine" and
       ($submit eq "Next" or $submit eq "Previous" or
	$submit eq "Date" or $submit eq "Sender" or
	$submit eq "Subject" or $submit eq "SA Score"))) { viewQuarantine(); }

# Whitelist Menu
elsif($submit eq "Whitelist") { viewWhitelist(); }
elsif($submit eq "Update" and $source eq "viewWhitelist") { updateWhitelist(); }

# Blacklist Menu
elsif($submit eq "Blacklist") { viewBlacklist(); }
elsif($submit eq "Update" and $source eq "viewBlacklist") { updateBlacklist(); }

# VirtUser Menu
elsif($submit eq "Email Aliases") { viewEmailAliases(); }
elsif($submit eq "Add New Alias" or $submit eq "Save Alias") { addEmailAlias(); }
elsif($submit eq "Edit" or $submit eq "Change Default Alias") { editEmailAlias(); }
elsif($submit eq "Delete") { delEmailAlias(); }

# Preferences Menu
elsif($submit eq "Preferences") { viewPreferences(); }
elsif($delaliasbutton) { delPrefsAlias(); }
elsif($submit eq "Add Aliased Account") { addPrefsAlias(); }
elsif($submit eq "Update Preferences") { updatePreferences(); }

# Logout
elsif($submit eq "Logout") { logout(); }

# Cookie logins won't have a $submit
elsif(!$submit) { viewQuarantine(); }

# Unknown
else { printError("Invalid submission"); }

# Log user actions
sub logAction {
    return unless defined $USERLOG;
    my $message = shift;
    my $logline = localtime(time)."\t$username\@$DEFAULT_DOMAIN ($identity) $message\n";
    if(open(LOGFILE, ">>$USERLOG")) {
        if(flock(LOGFILE,2)) {
            print LOGFILE $logline;
        }
        close(LOGFILE);
    } else {
        warn "Error opening logfile: $!\n";
    }
}

# Authenticate user
sub verifyLogin {
    #
    # Example of PAM authentication via spasm's authuser program:
    #
    my $rc = system($SPASM_AUTH_APP, $username, $password);
    if($rc != 0) { printLoginScreen("Invalid username or password"); }
    $identity = "$username\@$DEFAULT_DOMAIN";

    #
    # Example of multi-ISP authentication via LDAP:
    #
#    if($ISPS{$isp}) { $DEFAULT_DOMAIN = $ISPS{$isp}{domain}; }
#    else { printLoginScreen("Invalid ISP"); }
#
#    my $ldap = Net::LDAP->new($LDAP_SERVER, port=>$LDAP_PORT);
#    unless($ldap) { printLoginScreen("Error connecting to LDAP server: $@"); }
#    my $result = $ldap->bind("uid=$username,ou=People,dc=$isp,dc=com",password=>"$password");
#    if($result->code == 49) { printLoginScreen("Invalid username or password"); }
#    elsif($result->code) { printLoginScreen("Error binding to LDAP server: ", $result->error); }
#    $ldap->unbind();
#    $identity = "$username\@$ISPS{$isp}{domain}";

    #
    # Example of virtual domain authentication via LDAP:
    #
#    if($virtemail) { $identity = $virtemail; }
#    elsif($virtdomain) { $identity = $virtdomain; }
#    my $ldap = Net::LDAP->new($LDAP_SERVER, port=>$LDAP_PORT);
#    unless($ldap) { printLoginScreen("Error connecting to LDAP server: $@"); }
#    my $result = $ldap->bind("uid=$username,ou=People,dc=$isp,dc=com",password=>"$password");
#    if($result->code == 49) { printLoginScreen("Invalid username or password"); }
#    elsif($result->code) { printLoginScreen("Error binding to LDAP server: ", $result->error); }
#
#    # Get list of domains for user
#    $result = $ldap->search(base=>"uid=$username,ou=People,dc=$isp,dc=com",scope=>"base",filter=>"(virtualDomain=*)");
#    if($result->code) { printMenulessError("Error searching for username: ", $result->error); }
#    if(my $entry = $result->entry(0)) {
#        @domains = $entry->get_value('virtualDomain');
#    }
#
#    # Are we currently working on a domain? Verify it!
#    if($virtdomain) {
#        $result = $ldap->search(base=>"uid=$username,ou=People,dc=$isp,dc=com",scope=>"base",filter=>"(virtualDomain=$virtdomain)");
#        unless($result->entry(0)) { $ldap->unbind(); printMenulessError("You do not have access to $virtdomain"); }
#    } else {
#        $result = $ldap->search(base=>"uid=$username,ou=People,dc=$isp,dc=com",scope=>"base",filter=>"(virtualDomain=*)");
#        unless($result->entries) { $ldap->unbind(); printMenulessError("You do not have access to any domains"); }
#    }
#    
#    if($virtemail) {
#        unless($virtdomain) { printMenulessError("No domain specified"); }
#        unless(verifyEmail($virtemail)) { printMenulessError("You do not have access to $virtemail"); }
#    }
#    $ldap->unbind();

    #
    # Example of mixed virtual domain and multi-ISP authentication via LDAP:
    #
#    if($ISPS{$isp}) { $DEFAULT_DOMAIN = $ISPS{$isp}{domain}; }
#    else { printLoginScreen("Invalid ISP"); }
#
#    my $ldap = Net::LDAP->new($LDAP_SERVER, port=>$LDAP_PORT);
#    unless($ldap) { printLoginScreen("Error connecting to LDAP server: $@"); }
#    my $result = $ldap->bind("uid=$username,ou=People,dc=$isp,dc=com",password=>"$password");
#    if($result->code == 49) { printLoginScreen("Invalid username or password"); }
#    elsif($result->code) { printLoginScreen("Error binding to LDAP server: ", $result->error); }
#    # Get list of domains for user
#    $result = $ldap->search(base=>"uid=$username,ou=People,dc=$isp,dc=com",scope=>"base",filter=>"(virtualDomain=*)");
#    if($result->code) { printLoginScreen("Error searching for username: ", $result->error); }
#    if(my $entry = $result->entry(0)) {
#        @domains = $entry->get_value('virtualDomain');
#        push @domains, "$username\@$ISPS{$isp}{domain}";
#        unless($virtdomain) {
#            my $defaultdomainaccount = `$SPASM_CGI_APP lookup_prefs $username\@$DEFAULT_DOMAIN defaultdomainaccount`;
#            chomp $defaultdomainaccount;
#            if($defaultdomainaccount) { $virtdomain = $defaultdomainaccount; }
#            else { $virtdomain = "$username\@$ISPS{$isp}{domain}"; }
#        }
#    }
#    
#    # Are we currently working on a domain? Verify it!
#    if($virtdomain and $virtdomain ne "$username\@$DEFAULT_DOMAIN") {
#        $result = $ldap->search(base=>"uid=$username,ou=People,dc=$isp,dc=com",scope=>"base",filter=>"(virtualDomain=$virtdomain)");
#        unless($result->entry(0)) { $ldap->unbind(); printMenulessError("You do not have access to $virtdomain"); }
#        # Load %addresses
#        my $enable_virtual_addresses = `$SPASM_CGI_APP lookup_prefs $virtdomain enable_virtual_addresses`;
#        chomp $enable_virtual_addresses;
#        if($enable_virtual_addresses or $submit eq "Email Aliases") {
#            $result = $ldap->search(base=>"$LDAP_VIRTDN", scope=>"one",filter=>"(virtEmail=*\@$virtdomain)");
#            for my $entry ($result->sorted("virtEmail")) {
#                my $ve = $entry->get_value('virtEmail') || '';
#                 my $re = $entry->get_value('realEmail') || '';
#                if($ve and $re) { $addresses{$ve} = $re; }
#             }
#        }
#    }
#    # Are we currently working on a virtual email address? Verify it!
#    if($virtemail eq "Global" or $virtdomain eq "$username\@$DEFAULT_DOMAIN") { $virtemail = ""; }
#    if($virtemail) {
#        if($virtdomain =~ /@/) { printMenulessError("'$virtdomain' is not a valid domain!"); }
#        unless($virtdomain) { printMenulessError("No domain specified"); }
#        unless(verifyEmail($virtemail)) { $virtemail = ""; }
#    }
#    # Set the identity
#    if($virtemail) { $identity = $virtemail; }
#    elsif($virtdomain) { $identity = $virtdomain; }
#    else { $identity = "$username\@$DEFAULT_DOMAIN"; }
#
#    $ldap->unbind();

    # Deny login if the account is aliased
    my $alias = `$SPASM_CGI_APP lookup_prefs $username\@$DEFAULT_DOMAIN alias`;
    chomp $alias;
    if($alias) { printLoginScreen("That account is currently aliased to $alias"); }

    my $setcookie = `$SPASM_CGI_APP lookup_prefs $username\@$DEFAULT_DOMAIN setcookie`;
    chomp $setcookie;
    if($setcookie eq "yes") {
        $COOKIE_ID = $cgi->cookie(-name=>'spasmLoginID', -value=>"$username", -expires=>'+7d', -path=>"$SCRIPT");
        $COOKIE_PASS = $cgi->cookie(-name=>'spasmLoginPass', -value=>"$password", -expires=>'+7d', -path=>"$SCRIPT");
        if(%ISPS) {
            $COOKIE_ISP = $cgi->cookie(-name=>'spasmLoginISP', -value=>"$isp", -expires=>'+7d', -path=>"$SCRIPT");
        }
    }

    logAction("verified");
}

sub verifyAccount {
    my $aliasusername = shift;
    my $aliaspassword = shift;
    my $aliasisp = %ISPS?shift:"";

    my $rc = system($SPASM_AUTH_APP, $aliasusername, $aliaspassword);
    return !$rc;
}

sub readOptions {
    my $h1_font = `$SPASM_CGI_APP lookup_prefs $username\@$DEFAULT_DOMAIN h1_font`;
    chomp $h1_font;
    unless($h1_font) { $h1_font = "13pt"; }
    my $b_font = `$SPASM_CGI_APP lookup_prefs $username\@$DEFAULT_DOMAIN b_font`;
    chomp $b_font;
    unless($b_font) { $b_font = "12pt"; }

    $CSS = <<EOF;
      BODY { background: white; color: black; font-family: "times new roman", serif; font-size: $b_font; }
      H1 { font-size: $h1_font; text-align: center; }
      H2 { font-size: 10pt; text-align: center; }
      IMG { border: none; }
      TABLE { margin-left: auto; margin-right: auto; }
      FORM { margin-bottom: 0; }
      A:link { text-decoration: none; color: #004080; }
      A:visited { text-decoration: none; color: #004080; }
      A:hover { text-decoration: none; color: #990000; background: #ddddff; }                                                                          
      A:active { text-decoration: none; color: #004080; }                                                                                              
EOF

}

#
# The actual functions that do the real work:
#

# Logout routine
sub logout {
    if($COOKIE_ID) { $COOKIE_ID = $cgi->cookie(-name=>'spasmLoginID', -value=>"", -expires=>'0', -path=>"$SCRIPT"); }
    if($COOKIE_PASS) { $COOKIE_PASS = $cgi->cookie(-name=>'spasmLoginPass', -value=>"", -expires=>'0', -path=>"$SCRIPT"); }
    if($COOKIE_ISP) { $COOKIE_ISP = $cgi->cookie(-name=>'spasmLoginISP', -value=>"", -expires=>'0', -path=>"$SCRIPT"); }
    printLoginScreen();
}

# Error routine
sub printError {
    printHeader();
    printMenuBar();
    print "<H1>@_</H1>";
    printFooter();
    exit;
}

# No-menu error routine
sub printMenulessError {
    printHeader();
    print "<H1>@_</H1>";
    printFooter();
    exit;
}

# Sort routine for blacklists/whitelists
sub sortRoutine {
    my $c = $a;
    my $d = $b;
    chomp $c;
    chomp $d;
    $c =~ s/\s*\/\/.*$//;
    $d =~ s/\s*\/\/.*$//;
    if($c =~ /^[\d\.\/]+$/ and $d =~ /^[\d\.\/]+$/) {
	# IP address and IP address
        $c =~ s/\/.*//;
        $d =~ s/\/.*//;
        my $e = unpack("N", pack("C4", split(/\./,$c)));
        my $f = unpack("N", pack("C4", split(/\./,$d)));
        $e <=> $f;
    } elsif($c =~ /^[\d\.\/]+$/ and $d !~ /^[\d\.\/]+$/) {
        # IP address and domain/email
        return -1;
    } elsif($c !~ /^[\d\.\/]+$/ and $d =~ /^[\d\.\/]+$/) {
        # domain/email and IP address
        return 1;
    } elsif($c =~ /@/ or $d =~ /@/) {
	# domain/email and domain/email
        my $cu = "";
        my $cd = "";
	my $du = "";
	my $dd = "";                                                                                                                                    
        if($c =~ /@/) {
            if($c =~ /^(.+)?@(\S+)/) { $cu = $1; $cd = $2; }
            elsif($c =~ /^@(\S+)/) { $cd = $2; }
        } else { $cd = $c; }
        if($d =~ /@/) {
            if($d =~ /^(.+)?@(\S+)/) { $du = $1; $dd = $2; }
            elsif($d =~ /^@(\S+)/) { $dd = $2; }
        } else { $dd = $d; }
        
        if($cu and $du) { ($cd cmp $dd) or ($cu cmp $du); }
        else { $cd cmp $dd; }
    } else {
        $c cmp $d;
    }
}

# Multi-sort routine for sorting log/quarantine hashes
sub multiSortHash {
    my $type = shift;
    my $hash = shift;

    if($type eq "DateDown") {
	return ($hash->{$b}{timestamp} <=> $hash->{$a}{timestamp});
    }
    if($type eq "DateUp") {
	return ($hash->{$a}{timestamp} <=> $hash->{$b}{timestamp});
    }
    if($type eq "SA Score") {
        return ($hash->{$b}{score} <=> $hash->{$a}{score});
    }
    if($type eq "Filter") {
	return ($hash->{$a}{filter} cmp $hash->{$b}{filter});
    }
    if($type eq "StatusUp") {
	return ($hash->{$a}{status} <=> $hash->{$b}{status});
    }
    if($type eq "StatusDown") {
	return ($hash->{$b}{status} <=> $hash->{$a}{status});
    }
    if($type eq "Recipient") {
	return (lc($hash->{$a}{recipient}) cmp lc($hash->{$b}{recipient}));
    }
    if($type eq "Subject") {
	return (lc($hash->{$a}{subject}) cmp lc($hash->{$b}{subject}));
    }
    if($type eq "Sender") {
	my $c = $hash->{$a}{sender};
	chomp $c;
	my $d = $hash->{$b}{sender};
	chomp $d;
	$c =~ s/\s*\/\/.*$//;
	$d =~ s/\s*\/\/.*$//;
	if($c =~ /^[\d\.\/]+$/ and $d =~ /^[\d\.\/]+$/) {
	    $c =~ s/\/.*//;
	    $d =~ s/\/.*//;
	    my $e = unpack("N", pack("C4", split(/\./,$c)));
	    my $f = unpack("N", pack("C4", split(/\./,$d)));
	    return ($e <=> $f);
	} elsif($c =~ /@/ or $d =~ /@/) {
	    $c =~ /^(.+)?@(\S+)/;
	    my $cu = $1; my $cd = $2;
	    $d =~ /^(.+)?@(\S+)/;
	    my $du = $1; my $dd = $2;
            if(defined($cd) and defined($dd)) { return (($cd cmp $dd) or ($cu cmp $du)); }
            elsif(defined($cd)) { return ($cd cmp $du); }
            elsif(defined($dd)) { return ($cu cmp $dd); }
            else { return ($cu cmp $du); }
	} else { return($c cmp $d); }
    }
    # Just in case...
    return ($a cmp $b);
}

# Print an HTML header
sub printHeader {
    print "Content-type: text/html\n";
    if($COOKIE_ID) { print "Set-Cookie: $COOKIE_ID\n"; }
    if($COOKIE_PASS) { print "Set-Cookie: $COOKIE_PASS\n"; }
    if($COOKIE_ISP) { print "Set-Cookie: $COOKIE_ISP\n"; }
    print "\n";
    print <<EOF;
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01//EN" "http://www.w3.org/TR/html4/strict.dtd">
<HTML>
  <HEAD>
    <TITLE>spasm account administrator - $identity (v$VERSION)</TITLE>
    <META http-equiv="Content-Type" content="text/html; charset=utf-8">
    <STYLE type="text/css">
$CSS
    </STYLE>
    <SCRIPT type="text/javascript">
      function openHelpWindow(filter, filterHelp) {
        helpWindow=window.open('', 'SpasmHelp', 'width=600,height=200,menubar=no,resizable=no');
        helpWindow.focus();
        helpWindow.document.open();
        helpWindow.document.write('<HTML><HEAD><TITLE>Spasm Help</TITLE></HEAD><BODY style=\\'font-family: "times new roman", serif; font-size: 12pt;\\'>');
        helpWindow.document.write('<center><h1 style=\\'font-family: "times new roman", serif; font-size: 15pt; color: #990000;\\'>'+filter+'</h1></center><p>');
        helpWindow.document.write(filterHelp);
        helpWindow.document.write('</BODY></HTML>');
        helpWindow.document.close();
      }     
      function selectAll() {
        mydoc=document.myform;
        numElements=mydoc.elements.length;
        for(var i=0;i<numElements;i++) {
          if(mydoc.elements[i].type == 'checkbox') {
            mydoc.elements[i].checked = mydoc.selectUnselect.checked;
          }
        }
      }
      function checkAll() {
        mydoc=document.myform;
        numElements=mydoc.elements.length;
        for(var i=0;i<numElements;i++) {
          if(mydoc.elements[i].type == 'checkbox') {
            mydoc.elements[i].checked = 1;
            }
        }
      }
      function uncheckAll() {
        mydoc=document.myform;
        numElements=mydoc.elements.length;
        for(var i=0;i<numElements;i++) {
          if(mydoc.elements[i].type == 'checkbox') {
            mydoc.elements[i].checked = 0;
            }
        }
      }
      function loginFocus() {
        document.loginForm.elements[0].focus();
      }
      function passwordFocus() {
        document.loginForm.elements[1].focus();
      }
    </SCRIPT>
  </HEAD>
EOF
}

# Print an HTML header for the filter page
sub printFilterHeader {
    my $numFilters = scalar(@SF);
    my $farray = "";
    for (@SF) { $farray .= "'$_',"; }
    chop $farray;
    print "Content-type: text/html\n";
    if($COOKIE_ID) { print "Set-Cookie: $COOKIE_ID\n"; }
    if($COOKIE_PASS) { print "Set-Cookie: $COOKIE_PASS\n"; }
    if($COOKIE_ISP) { print "Set-Cookie: $COOKIE_ISP\n"; }
    print "\n";
    print <<EOF;
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01//EN" "http://www.w3.org/TR/html4/strict.dtd">
<HTML>
  <HEAD>
    <TITLE>spasm account administrator - $identity (v$VERSION)</TITLE>
    <META http-equiv="Content-Type" content="text/html; charset=utf-8">
    <STYLE type="text/css">
$CSS
    </STYLE>
    <SCRIPT type="text/javascript">
      function openHelpWindow(filter, filterHelp) {
        helpWindow=window.open('', 'SpasmHelp', 'width=600,height=200,menubar=no,resizable=no');
        helpWindow.focus();
        helpWindow.document.open();
        helpWindow.document.write('<HTML><HEAD><TITLE>Spasm Help</TITLE></HEAD><BODY style=\\'font-family: "times new roman", serif; font-size: 12pt;\\'>');
        helpWindow.document.write('<center><h1 style=\\'font-family: "times new roman", serif; font-size: 15pt; color: #990000;\\'>'+filter+'</h1></center><p>');
        helpWindow.document.write(filterHelp);
        helpWindow.document.write('</BODY></HTML>');
        helpWindow.document.close();
      }     
      function checkRadioButton(myradio) {
        fArray = [$farray];

        if(myradio == 5) { document.getElementById("customTable").style.visibility = "visible"; }
        else { document.getElementById("customTable").style.visibility = "hidden"; }

        if(myradio == 5) {
        }
        else if(myradio == 4) {
          for(var i=0; i<$numFilters; i++) {
            obj = document.getElementsByName(fArray[i]);
            if(fArray[i] == "SA_REQUIRED_HITS") { obj[0].value = "4"; }
            else if(fArray[i] == "SF_WHITELIST_ONLY") {
              obj[0].checked = 1;
              obj = document.GetElementsByName("logopt-" + fArray[i]);
              obj[0].checked = 0;
            } else {
              obj[2].checked = 1;
              obj = document.getElementsByName("logopt-" + fArray[i]);
              obj[0].checked = 1; }
          }
        } else {
          if(myradio >= 0) {
            for(var i=0; i<$numFilters; i++) {
              obj = document.getElementsByName(fArray[i]);
              obj[0].checked = 1;
              if(fArray[i] != "SA_REQUIRED_HITS") {
                obj = document.getElementsByName("logopt-" + fArray[i]);
                obj[0].checked = 0;
              }
            }
          }
          if(myradio >= 1) {
            document.myform.SF_VIRUS_SCAN[2].checked = 1;
            document.myform.SF_PERSONAL_BLACKLIST[2].checked = 1;
            document.myform.SF_SBL[2].checked = 1;
            document.myform.SF_SORBS[2].checked = 1;
          }
          if(myradio >= 2) {
            document.myform.SF_SPAMTRAP[2].checked = 1;
            document.myform.SF_LOCALFILTER[2].checked = 1;
            document.myform.SF_SPAMASSASSIN[2].checked = 1;
            document.myform.SA_REQUIRED_HITS.value = "5";
            document.myform.SF_SPAMCOP[2].checked = 1;
          }
          if(myradio >= 3) {
            document.myform.SF_RDNS[2].checked = 1;
            document.myform.SF_HELO[2].checked = 1;
            document.myform.SF_MAIL_HOST[2].checked = 1;
            document.myform.SA_REQUIRED_HITS.value = "4";
            document.myform.SF_POSTMASTERRFC[2].checked = 1;
            document.myform.SF_DSNRFC[2].checked = 1;
            document.myform.SF_ABUSERFC[2].checked = 1;
            document.myform.SF_WHOISRFC[2].checked = 1;
          }
        }
      }
    </SCRIPT>
  </HEAD>
EOF
}

# Print an HTML footer
sub printFooter {
    print <<EOF;
    </FORM>
  </BODY>
</HTML>
EOF
}

# Print the menu bar
sub printMenuBar {
    my $domainoptions = "";
    my $addressoptions = "";

    if(@domains) {
	$domainoptions = "<select name='virtdomain'>";
        for (@domains) {
            if($_ eq $virtdomain) { $domainoptions .= "<option selected value='$_'>$_"; }
            else { $domainoptions .= "<option value='$_'>$_"; }
        }
        $domainoptions .= "</select>";

        if($virtdomain eq "$username\@$DEFAULT_DOMAIN") {
            $addressoptions = "&nbsp;";
        } else {
            $addressoptions = "<select name='virtemail'>";
            $addressoptions .= "<option value='Global'>[Global]</option>";
            for (sort(keys %addresses)) {
                if($_ =~ /^\@/) { next; }
                if($_ eq $virtemail) { $addressoptions .= "<option selected value='$_'>$_</option>"; }
                else { $addressoptions .= "<option value='$_'>$_</option>"; }
            }
            $addressoptions .= "</select>";
        }
    }

    print <<EOF;
  <BODY>
    <FORM action='$SCRIPT' method='POST' name='myform'>
      <P>
      <input type=hidden name=username value="$username">
      <input type=hidden name=password value="$password">
      <input type=hidden name=isp value="$isp">
      <TABLE width='100%'>
        <COL width="33%">
        <COL width="33%">
        <COL width="33%">
        <TR>
          <TD align=left><img src="$GFX_URL/spasm-logo.png" alt="Spasm">
          <TD align=center>
	    <TABLE width='100%'>
	      <TR><TD align=center>$domainoptions</TR>
	      <TR><TD align=center>$addressoptions</TR>
	    </TABLE>
          <TD align=right>
            <TABLE width='100%'>
              <TR><TD align=right><a href='http://www.theasylum.org/spasm'>Spasm Homepage</a></TR>
              <TR><TD align=right>&nbsp;</TR>
              <TR><TD>&nbsp;</TR>
              <TR><TD align=right><a href='/spasmdoc'>How To Use Spasm</a></TR>
              <TR><TD align=right><a href='$SCRIPT?submit=Logout'>Logout</a></TR>
            </TABLE>
        </TR>
      </TABLE>
      <TABLE>
        <TR>
EOF
    if(@domains and $virtdomain ne "$username\@$DEFAULT_DOMAIN") { print "<TD><input type=submit name=submit value='Email Aliases'>"; }
    print <<EOF;
          <TD><input type=submit name=submit value="Mail Filters">
          <TD><input type=submit name=submit value="Whitelist">
          <TD><input type=submit name=submit value="Blacklist">
          <TD><input type=submit name=submit value="Log">
          <TD><input type=submit name=submit value="Quarantine">
	  <TD><input type=submit name=submit value="Preferences">
        </TR>                                                                                                      
      </TABLE>
      <HR>
EOF
}

# Print the login screen
sub printLoginScreen {
    my $responsemsg = shift || '';
    my $selectbox;
    my $focusOption;
    if(%ISPS) {
	for my $key (sort { $ISPS{$a}{order} <=> $ISPS{$b}{order} } keys(%ISPS)) {
            if($key eq $isp) { $selectbox .= "<option value='$key' selected>$key</option>"; }
            else { $selectbox .= "<option value='$key'>$key</option>"; }
	}
    }
    if($username) { $focusOption = "passwordFocus()"; }
    else { $focusOption = "loginFocus()"; }
    printHeader();
    print <<EOF;
  <BODY onLoad="$focusOption">
    <FORM name='loginForm' action='$SCRIPT' method='POST'>
      <TABLE style="margin-left: auto; margin-right: auto; background: #ddddff;" border=1>
        <TR><TD>Username:<TD><input type=text size=16 name=username value="$username"></TR>
        <TR><TD>Password:<TD><input type=password size=16 name=password value=""></TR>
EOF
    if($selectbox) {
	print "<TR><TD align=center>ISP:<TD align=center><select name=isp>$selectbox</select></TR>\n";
    }
    print <<EOF;
      </TABLE>
      <P style="text-align: center">
        <input type=submit name=submit value="Login">
      </P>
      <P style="text-align: center">
        $responsemsg
      </P>
EOF
    printFooter();
    exit;
}

# View mail filters screen
sub viewMailFilters {
    printFilterHeader();
    printMenuBar();
    # Get sort option
    my $filtersort = `$SPASM_CGI_APP lookup_prefs $username\@$DEFAULT_DOMAIN filtersort`;
    chomp $filtersort;

    my %FSTATUS = ();
    my $FGROUP = 5;
    my $flag;
    for my $filter (@SF) {
        $FSTATUS{$filter} = `$SPASM_CGI_APP lookup_filter $identity $filter`;
        chomp $FSTATUS{$filter};
    }
    print <<EOF;              
      <P>
      <input type=hidden name=virtdomain value="$virtdomain">
      <input type=hidden name=virtemail value="$virtemail">
      <input type=hidden name=source value="viewMailFilters">
      <H1>Basic Settings</H1>
      <P>
      <TABLE>
        <TR><TD>
EOF
    
    # Check for "Off" Setting
    $flag = 1;
    for(@SF) {
        if($_ eq "SA_REQUIRED_HITS") { next; }
        if($FSTATUS{$_} != 0) { $flag = 0; }
    }
    if($flag) { $FGROUP = 0; }
    
    # Check for "All" Setting
    $flag = 1;
    for(@SF) {
        if($_ eq "SA_REQUIRED_HITS") {
            if($FSTATUS{$_} != 4) { $flag = 0; }
            next;
        }
        if($_ eq "SF_WHITELIST_ONLY") {
            if($FSTATUS{$_} != 0) { $flag = 0; }
            next;
        }
        if($FSTATUS{$_} != 2) { $flag = 0; }
    }
    if($flag) { $FGROUP = 4; }
    
    # Check for "Low" Setting
    $flag = 1;
    for(@SF) {
        if($_ eq "SA_REQUIRED_HITS") { next; }
        if($FSTATUS{$_} != $FG_LOW{$_}) { $flag = 0; }
    }
    if($flag) { $FGROUP = 1; }

    # Check for "Medium" Setting
    $flag = 1;
    for(@SF) {
        if($FSTATUS{$_} != $FG_MED{$_}) { $flag = 0; }
    }
    if($flag) { $FGROUP = 2; }

    # Check for "High" Setting
    $flag = 1;
    for(@SF) {
        if($FSTATUS{$_} != $FG_HIGH{$_}) { $flag = 0; }
    }
    if($flag) { $FGROUP = 3; }

    for(0..5) {
        if($_ == $FGROUP) {
            print "<input type=radio name=newfg value='$_' onClick='checkRadioButton(this.value)' checked>$FILTERGROUP[$_]\n";
        } else {
            print "<input type=radio name=newfg value='$_' onClick='checkRadioButton(this.value)'>$FILTERGROUP[$_]\n";
        }
    }
    
    my $visibleCustom;
    if($FGROUP == 5) { $visibleCustom = "visible"; }
    else { $visibleCustom = "hidden"; }

    print <<EOF;
        </TR>
      </TABLE>
      <P>
      <TABLE>
        <TR><TD align=center><input type=submit name=submit value="Update"></TR>
      </TABLE>
      <DIV Id="customTable" class="customTable" style="visibility: $visibleCustom;">
      <P>
      <H1>Custom Settings</H1>
      <H2 style="color: #004080">Warning: For Advanced Users Only</H2>
      <P>
      <TABLE frame=border rules=rows style="background:#ddddff;">
        <TR><TH>Filter<TH>Off<TH>Tag<TH>Quarantine<TH>Reject<TH>Log</TR>
EOF

    my @filters;
    if($filtersort eq "Alphabetical") {
        @filters = sort(@SF);
    } else {
        @filters = @SF;
    }
    for my $filter (@filters) {
        if($filter eq "SA_REQUIRED_HITS") { next; } # This comes later
        my $fstatus = $FSTATUS{$filter};
	my $helplink = "<a href=\"#\" onClick=\"openHelpWindow('$filter', '$FILTER_HELP{$filter}'); return false;\">$filter</a>";
        print "<TR>";
        print "<TD>$helplink";
        if($fstatus == 0 or $fstatus == 8) {
            print "<TD align=center><input type=radio name=$filter value='0' checked>";
            print "<TD align=center><input type=radio name=$filter value='1'>";
            print "<TD align=center><input type=radio name=$filter value='2'>";
            print "<TD align=center><input type=radio name=$filter value='4'>";
            if($fstatus & 8) {
                print "<TD align=center><input type=checkbox name=logopt-$filter value='8' checked>";
            } else {
                print "<TD align=center><input type=checkbox name=logopt-$filter value='8'>";
            }
        } elsif($fstatus & 1) {
            print "<TD align=center><input type=radio name=$filter value='0'>";
            print "<TD align=center><input type=radio name=$filter value='1' checked>";
            print "<TD align=center><input type=radio name=$filter value='2'>";
            print "<TD align=center><input type=radio name=$filter value='4'>";
            if($fstatus & 8) {
                print "<TD align=center><input type=checkbox name=logopt-$filter value='8' checked>";
            } else {
                print "<TD align=center><input type=checkbox name=logopt-$filter value='8'>";
            }
        } elsif($fstatus & 2) {                                           
            print "<TD align=center><input type=radio name=$filter value='0'>";
            print "<TD align=center><input type=radio name=$filter value='1'>";
            print "<TD align=center><input type=radio name=$filter value='2' checked>";
            print "<TD align=center><input type=radio name=$filter value='4'>";
            if($fstatus & 8) {
                print "<TD align=center><input type=checkbox name=logopt-$filter value='8' checked>";
            } else {
                print "<TD align=center><input type=checkbox name=logopt-$filter value='8'>";
            }
        } elsif($fstatus & 4) {                                           
            print "<TD align=center><input type=radio name=$filter value='0'>";
            print "<TD align=center><input type=radio name=$filter value='1'>";
            print "<TD align=center><input type=radio name=$filter value='2'>";
            print "<TD align=center><input type=radio name=$filter value='4' checked>";
            if($fstatus & 8) {
                print "<TD align=center><input type=checkbox name=logopt-$filter value='8' checked>";
            } else {
                print "<TD align=center><input type=checkbox name=logopt-$filter value='8'>";
            }
        } else { warn "Unexpected fstatus for $identity: $fstatus\n"; }
        print "</TR>\n";
        if($filter eq "SF_SPAMASSASSIN") {
            # Keep the SA_REQUIRED_HITS filter after SF_SPAMASSASSIN
            $filter = "SA_REQUIRED_HITS";
            my $fstatus = $FSTATUS{$filter};
            my $helplink = "<a href=\"#\" onClick=\"openHelpWindow('$filter', '$FILTER_HELP{$filter}'); return false;\">$filter</a>";
            print "<TR><TD>$helplink";
            print "<TD><input type=text size=3 maxlength=3 name='SA_REQUIRED_HITS' value='$fstatus'></tr>";
        }
    }
    
    print <<EOF;
      </TABLE>
      <P>
      <TABLE>
        <TR><TD align=center><input type=submit name=submit value="Update"></TR>
      </TABLE>
      </DIV>
EOF
    printFooter();
}

# Update mail filters
sub updateMailFilters {
    my $newfg = $cgi->param('newfg')?$cgi->param('newfg'):0;
    if($newfg == 0) {
        for my $filter(@SF) {
            system($SPASM_CGI_APP, "set_filter", "$identity", "$filter", "0");
        }
    } elsif($newfg == 1) {
        for my $filter(@SF) {
            my $fstatus = $FG_LOW{$filter};
            system($SPASM_CGI_APP, "set_filter", "$identity", "$filter", "$fstatus");
        }
    } elsif($newfg == 2) {
        for my $filter(@SF) {
            my $fstatus = $FG_MED{$filter};
            system($SPASM_CGI_APP, "set_filter", "$identity", "$filter", "$fstatus");
        }
    } elsif($newfg == 3) {
        for my $filter(@SF) {
            my $fstatus = $FG_HIGH{$filter};
            system($SPASM_CGI_APP, "set_filter", "$identity", "$filter", "$fstatus");
        }
    } elsif($newfg == 4) {
        for my $filter(@SF) {
            my $fstatus = 2;
            if($filter eq "SA_REQUIRED_HITS") { $fstatus = 4; }
            if($filter eq "SF_WHITELIST_ONLY") { $fstatus = 0; }
            system($SPASM_CGI_APP, "set_filter", "$identity", "$filter", "$fstatus");
        }
    } else {
        for my $filter(@SF) {
            my $fstatus = $cgi->param($filter)?$cgi->param($filter):0;
	    my $lstatus = $cgi->param("logopt-$filter")?$cgi->param("logopt-$filter"):0;
	    my $status = int($fstatus) | int($lstatus);
            system($SPASM_CGI_APP, "set_filter", "$identity", "$filter", "$status");
        }
    }    
    printHeader();
    printMenuBar();
    print <<EOF;
    <H1>Mail Filters Updated</H1>
EOF
    printFooter();
    logAction("updated mail filters");
}

# View log screen
sub viewLog {
    if($virtemail and $identity =~ /@(.+)$/) { printError("Logs must be viewed in Global context"); }

    # Get sort variable, default is DateDown sorting
    my $sort = $cgi->param('sort')?$cgi->param('sort'):"DateDown";
    # Set sorting based on submit button, if necessary
    if($submit eq "Date" or $submit eq "Sender" or $submit eq "Filter" or $submit eq "Status" or $submit eq "Recipient") {
        if($submit eq "Date" and $sort eq "DateUp") { $sort = "DateDown"; }
        elsif($submit eq "Date" and $sort eq "DateDown") { $sort = "DateUp"; }
        elsif($submit eq "Date") { $sort = "DateDown"; } 
	elsif($submit eq "Status" and $sort eq "StatusUp") { $sort = "StatusDown"; }
	elsif($submit eq "Status" and $sort eq "StatusDown") { $sort = "StatusUp"; }
	elsif($submit eq "Status") { $sort = "StatusUp"; }
	else { $sort = $submit; }
    }   
    # If an invalid sort variable was passed, force DateDown sorting
    unless($sort eq "DateUp" or $sort eq "DateDown" or $sort eq "Sender" or
	   $sort eq "Filter" or $sort eq "StatusUp" or $sort eq "StatusDown" or $sort eq "Recipient") {
        $sort = "DateDown";
    }   
    # Initialise paging
    my $pageall = 0;
    my $pageby = `$SPASM_CGI_APP lookup_prefs $username\@$DEFAULT_DOMAIN pageby`;
    chomp $pageby;
    unless($pageby) { $pageby = 25; }
    if($pageby eq 'All') { $pageall = 1; }
    else { $pageby--; }
    my $totalnum = 0; 
    
    # Get logs
    my @logs = `$SPASM_CGI_APP view_log $identity ALL`;

    # Exit if the log doesn't exist
    unless(@logs) {
        printHeader();
        printMenuBar();
        print "<H1>Log Empty</H1>";
        printFooter();
        exit;
    }

    my $max_sender_length = `$SPASM_CGI_APP lookup_prefs $username\@$DEFAULT_DOMAIN max_sender_length`;
    chomp $max_sender_length;

    # Parse logs
    my %logHash;
    for my $entry (@logs) {
	chop $entry;
        $entry =~ s/</&lt\;/g;
        my @fields = split(/:/, $entry);
	$logHash{$fields[0]}{timestamp} = $fields[1];
        $logHash{$fields[0]}{date} = "$fields[2]:$fields[3]:$fields[4]";
	$logHash{$fields[0]}{recipient} = $fields[5];
        $logHash{$fields[0]}{sender} = $fields[6];
	if($max_sender_length) { $logHash{$fields[0]}{sender} =~ s/^(.{$max_sender_length}).*$/$1/; }
        $logHash{$fields[0]}{status} = defined($fields[11])?$fields[11]:0;
	my @filts = split(/\|/, $fields[10]);
	$logHash{$fields[0]}{filter} = $filts[0];
        $totalnum++;
    }

    # Complete the paging setup
    my $entrynum = $cgi->param('entrynum')?$cgi->param('entrynum'):1;
    my $reload = $cgi->param('reload')?$cgi->param('reload'):0;
    my $curentry = 0;
    my $startentry = 1;
    my $endentry;
    if($pageall) { $endentry = $totalnum; }
    else { $endentry = $startentry+$pageby; }
    my $next = 0;
    my $data = "";

    # Unless they want everything, handle Next and Previous buttons, as well as a page reload
    if(!$pageall) {
        if($submit eq "Next") {
            $startentry = $entrynum + $pageby + 1;
            $endentry = $startentry+$pageby;
        } elsif($submit eq "Previous" or $reload == 2) {
            $endentry = $entrynum-1;
            $startentry = $endentry - $pageby;
        } elsif($reload == 1) {
	    $startentry = $entrynum;
	    $endentry = $startentry + $pageby;
	}
    }

    # Start entries can't be < 1
    if($startentry < 1) { $startentry = 1; }

    # Date Formatting
    my $dateformat = `$SPASM_CGI_APP lookup_prefs $username\@$DEFAULT_DOMAIN dateformat`;
    chomp $dateformat;
    unless($dateformat) { $dateformat = "Day MM DD hh:mm:ss YYYY"; }

    # Sort the entries
    for my $key (sort { multiSortHash($sort, \%logHash) } keys %logHash) {
        $curentry++;
        if($curentry > $endentry) { $next = 1; last; }
        if($curentry < $startentry) { next; }

        my $formatted_date = "";
        my @date_fields = split(" ", $logHash{$key}{date});
        if($dateformat eq "Day MM DD hh:mm:ss YYYY") {
            $formatted_date = $logHash{$key}{date};
        }
        elsif($dateformat eq "MM DD hh:mm:ss YYYY") {
            $formatted_date = "$date_fields[1] $date_fields[2] $date_fields[3] $date_fields[4]";
        }
        elsif($dateformat eq "MM DD hh:mm:ss") {
            $formatted_date = "$date_fields[1] $date_fields[2] $date_fields[3]";
        }
        elsif($dateformat eq "MM DD") {
            $formatted_date = "$date_fields[1] $date_fields[2]";
        }
        else { printError("Invalid Date Format: $dateformat"); }

        $data .= "<TR><TD align=center><input type=checkbox name=logentry value='$key'><TD nowrap>$formatted_date<TD align=center nowrap>$logHash{$key}{sender}";
	if(@domains and $virtdomain !~ /@/) { $data .= "<TD align=center nowrap>$logHash{$key}{recipient}"; }
	$data .= "<TD align=left>$logHash{$key}{filter}<TD align=left>$STATUS[$logHash{$key}{status}]</TR>\n";
    }

    # End entry can't be greater than final current entry
    if($endentry > $curentry) { $endentry = $curentry; }

    unless($endentry) {
       printHeader();
       printMenuBar();
       print "<H1>Log Empty</H1>";
       printFooter();
       exit;
    }

    # Setup table headers
    my $theaders = "<TR><TH><input type=checkbox name=selectUnselect title='Check All/Clear All' onClick='javascript:selectAll()'><TH><input type=submit name=submit value='Date' title='Sort by Date' style='width: 100%'><TH><input type=submit name=submit value='Sender' title='Sort by Sender' style='width: 100%'>";
    if(@domains and $virtdomain !~ /@/) { $theaders .= "<TH><input type=submit name=submit value='Recipient' title='Sort by Recipient' style='width: 100%'>"; }
    $theaders .= "<TH><input type=submit name=submit value='Filter' title='Sort by Filter' style='width: 100%'><TH><input type=submit name=submit value='Status' title='Sort by Status' style='width: 100%'></TR>";

    # Print the page
    printHeader();
    printMenuBar();

    print <<EOF;
    <P>
    <input type=hidden name=virtdomain value="$virtdomain">
    <input type=hidden name=source value="viewLog">
    <input type=hidden name=sort value="$sort">
    <TABLE>
      <TR><TD align=center>
            <select name=selectMenu>
                <option value="View Selected Entries">View Selected Entries</option>
                <option value="View All Entries">View All Entries</option>
		<option value="Dump Log">Dump Log to Screen</option>
		<option value="Download Log">Download Log</option>
                <option value="Whitelist Selected Senders">Whitelist Selected Senders</option>
                <option value="Blacklist Selected Senders">Blacklist Selected Senders</option>
                <option value="Delete Selected Entries">Delete Selected Entries</option>
                <option value="Delete All Entries">Delete All Entries</option>
            </select>
          <TD align=center><input type=submit name=submit value="Submit">
      </TR>
    </TABLE>
    <p>
    <center>Viewing Entries $startentry - $endentry of $totalnum</center>
    <TABLE border=1 style="background:#ddddff;">
$theaders
$data
    <tr><td colspan=5 align=left bgcolor='#FFFFFF' style='border:none'><a href="#" onClick='javascript:checkAll(); return false;'>Check All</a> - <a href="#" onClick='javascript:uncheckAll(); return false;'>Clear All</a></tr>
    </TABLE>
    </P>
    <input type=hidden name=entrynum value='$startentry'>
    <input type=hidden name=endentry value='$endentry'>
    <center>
EOF
    # If we aren't showing everything, print Next/Previous buttons as required
    if(!$pageall) {
        if($curentry - ($pageby+1) > 1) {
            print "<input type=submit name=submit value='Previous'>";
        }   
        if($next) { 
            print "<input type=submit name=submit value='Next'>";
        }
    }
    print <<EOF;
    <p>
    <select name=selectMenu2>
      <option value="View Selected Entries">View Selected Entries</option>
      <option value="View All Entries">View All Entries</option>
      <option value="Dump Log">Dump Log to Screen</option>
      <option value="Download Log">Download Log</option>
      <option value="Whitelist Selected Senders">Whitelist Selected Senders</option>
      <option value="Blacklist Selected Senders">Blacklist Selected Senders</option>
      <option value="Delete Selected Entries">Delete Selected Entries</option>
      <option value="Delete All Entries">Delete All Entries</option>
    </select>
    <input type=submit name=submit2 value="Submit">
    </center>
EOF
    printFooter();
}

# Delete selected log entries
sub deleteLogEntry {
    my $deleted = 0;
    if($submit eq "Delete All Entries") {
	`$SPASM_CGI_APP delete_log $identity ALL`;
    } else {
	my @lfile = $cgi->param('logentry')?$cgi->param('logentry'):();
	unless(@lfile) { printError("No entries specified"); }
	for my $lfile (@lfile) {
	    `$SPASM_CGI_APP delete_log $identity $lfile`;
	    $deleted++;
	}
    }
    $cgi->param('reload', 1);
    my $startentry = $cgi->param('entrynum')?$cgi->param('entrynum'):1;
    my $endentry = $cgi->param('endentry')?$cgi->param('endentry'):1;
    if($deleted > 0 and $startentry != 1 and ($endentry - $deleted < $startentry)) {
	# We deleted the last entry[s] on the page. Go to previous page.
	$cgi->param('reload', 2);
    }
    if($submit eq "Delete All Entries") {
	logAction("deleted all log entries");
    } else {
	logAction("deleted $deleted log entries");
    }
    viewLog();
}

# View one or more full log entries
sub viewLogEntry {
    my @logEntries = ();
    my @logs = ();
    if($submit eq "View All Entries") {
	@logs = `$SPASM_CGI_APP view_log $identity ALL`;
    } else {
	@logEntries = $cgi->param('logentry')?$cgi->param('logentry'):();
	unless(@logEntries) { printError("No entries selected"); }
	for (@logEntries) {
	    my $entry = `$SPASM_CGI_APP view_log $identity $_`;
	    push @logs, $entry;
	}
    }
    printHeader();                                                                                                   
    printMenuBar();                                                                                                  
    for (@logs) {
	chomp;
        s/</&lt;/g;
        s/>/&gt;/g;
        my @fields = split(/:/, $_);
        $fields[10] =~ s/\|/<br>/g;
        $fields[11] = defined($fields[11])?$STATUS[$fields[11]]:"";
        print <<EOF;
    <TABLE border=1 width=75%>
      <TR><TH>Date<TD align=center>$fields[2]:$fields[3]:$fields[4]</TR>
      <TR><TH>Sender<TD align=center>$fields[6]</TR>
      <TR><TH>Hostname<TD align=center>$fields[7]</TR>
      <TR><TH>Host IP<TD align=center>$fields[8]</TR>
      <TR><TH>Host HELO<TD align=center>$fields[9]</TR>
      <TR><TH>Recipient<TD align=center>$fields[5]</TR>
      <TR><TH>Filters<TD align=center>$fields[10]</TR>
      <TR><TH>Status<TD align=center>$fields[11]</TR>
    </TABLE>
    <P>&nbsp;<P>
EOF
    }
    printFooter();
}

# Dump raw log to screen
sub dumpLog {
    my @logs = `$SPASM_CGI_APP view_log $identity ALL`;
    print "Content-type: text/plain\n";
    if($submit eq "Download Log") {
	print "Content-Disposition: attachment; filename=\"spasm.log\";\n\n";
    } else {
	print "\n";
    }
    for my $entry (@logs) {
	# Remove primary key from dump
	my ($key, $data) = split(/:/, $entry, 2);
	print $data;
    }
}

# View quarantine
sub viewQuarantine {
    if($virtemail and $identity =~ /@(.+)$/) { printError("Quarantine must be viewed in Global context"); }

    my $message = shift;
    $message = "" unless defined $message;
    my $sort = $cgi->param('sort')?$cgi->param('sort'):"DateDown";
    if($submit eq "Date" or $submit eq "Sender" or $submit eq "Subject" or $submit eq "SA Score") {
        if($submit eq "Date" and $sort eq "DateUp") { $sort = "DateDown"; }
        elsif($submit eq "Date" and $sort eq "DateDown") { $sort = "DateUp"; }
        elsif($submit eq "Date") { $sort = "DateDown"; }
        else { $sort = $submit; }
    }
    unless($sort eq "DateDown" or $sort eq "DateUp" or $sort eq "Subject" or $sort eq "Sender" or $sort eq "SA Score") {
	$sort = "DateDown";
    }
    my $pageall = 0;                                                                                                 
    my $pageby = `$SPASM_CGI_APP lookup_prefs $username\@$DEFAULT_DOMAIN pageby`;
    chomp $pageby;
    unless($pageby) { $pageby = 25; }
    if($pageby eq 'All') { $pageall = 1; }
    else { $pageby--; }
    my $totalnum = 0;

    my $max_subject_length = `$SPASM_CGI_APP lookup_prefs $username\@$DEFAULT_DOMAIN max_subject_length`;
    chomp $max_subject_length;

    my $max_sender_length = `$SPASM_CGI_APP lookup_prefs $username\@$DEFAULT_DOMAIN max_sender_length`;
    chomp $max_sender_length;

    my $view_sa_score = `$SPASM_CGI_APP lookup_prefs $username\@$DEFAULT_DOMAIN view_sa_score`;
    chomp $view_sa_score;

    unless(-d "$SPASM_QDIR/$identity") {
        printHeader();
        printMenuBar();
        print <<EOF;
  $message
  <H1>Quarantine is Empty</H1>
EOF
        printFooter();
        exit;
    }
    my %qHash;
    opendir(DIR, "$SPASM_QDIR/$identity") or
        printError("Unable to open quarantine directory: $!");
    while(my $file = readdir(DIR)) {
        my $sender = "";
        my $subject = "";
        if($file eq "." or $file eq "..") { next; }
        my @fstat = stat("$SPASM_QDIR/$identity/$file") or next;
        if($fstat[7] == 0) {
            next;
        }
        open(IF, "$SPASM_QDIR/$identity/$file") or                                      
            printError("Unable to open quarantine file: $!");;
        $file =~ /^(\d+)/;
        my $timestamp = $1;
	unless(defined $timestamp) { printError("Error parsing timestamp from quarantine file $file"); }
        $qHash{$file}{timestamp} = $timestamp;
	$qHash{$file}{virus} = 0;
        while(my $line = <IF>) {
            $line =~ s/[\r|\n]//g;
            if($line =~ /^X-spasm-sender: (.+)$/) {
                $qHash{$file}{sender} = $1;
                $qHash{$file}{sender} =~ s/</&lt;/g;
		if($max_sender_length) { $qHash{$file}{sender} =~ s/^(.{$max_sender_length}).*$/$1/; }
            } elsif($line =~ /^Subject: (.+)$/i) {
                $qHash{$file}{subject} = $1;
                $qHash{$file}{subject} =~ s/</&lt;/g;
		if($max_subject_length) { $qHash{$file}{subject} =~ s/^(.{$max_subject_length}).*$/$1/; }
            } elsif($line =~ /^X-spasm-date: (.+)$/) {
                $qHash{$file}{date} = $1;
	    } elsif($line =~ /^X-spasm-Virus:/) {
		$qHash{$file}{virus} = 1;
            } elsif($line =~ /^X-spasm-SA: (\d+\.\d+)/) {
                $qHash{$file}{score} = $1;
            } elsif($line =~ /^[\r|\n]*$/) { last; }
        }
        if(!$qHash{$file}{subject} or $qHash{$file}{subject}=~/^\s+$/) { $qHash{$file}{subject} = '&nbsp;'; }
        if(!$qHash{$file}{date}) { $qHash{$file}{date} = '&nbsp;'; }
	if(!$qHash{$file}{score}) { $qHash{$file}{score} = '&nbsp;'; }
        close(IF);
        $totalnum++;
    }
    closedir(DIR);

    my $entrynum = $cgi->param('entrynum')?$cgi->param('entrynum'):1;
    my $reload = $cgi->param('reload')?$cgi->param('reload'):0;
    my $curentry = 0;
    my $startentry = 1;
    my $endentry;
    if($pageall) { $endentry = $totalnum; }
    else { $endentry = $startentry+$pageby; }
    my $next = 0;
    my $data = "";

    if(!$pageall) {
        if($submit eq "Next") {
            $startentry = $entrynum + $pageby + 1;
            $endentry = $startentry+$pageby;
        } elsif($submit eq "Previous" or $reload == 2) {
            $endentry = $entrynum-1;
            $startentry = $endentry - $pageby;
        } elsif($reload == 1) {
            $startentry = $entrynum;
            $endentry = $startentry + $pageby;
        }
    }

    if($startentry < 1) { $startentry = 1; }

    # Date Formatting
    my $dateformat = `$SPASM_CGI_APP lookup_prefs $username\@$DEFAULT_DOMAIN dateformat`;
    chomp $dateformat;
    unless($dateformat) { $dateformat = "Day MM DD hh:mm:ss YYYY"; }

    # Sort the entries
    for my $key (sort { multiSortHash($sort, \%qHash) } keys %qHash) {
        $curentry++;
        if($curentry > $endentry) { $next = 1; last; }
        if($curentry < $startentry) { next; }
	my $virusBackground = $qHash{$key}{virus}?'style="background: #ff6060"':'';

	my $formatted_date = "";
	my @date_fields = split(" ", $qHash{$key}{date});
	if($dateformat eq "Day MM DD hh:mm:ss YYYY") {
	    $formatted_date = $qHash{$key}{date};
	}
	elsif($dateformat eq "MM DD hh:mm:ss YYYY") {
	    $formatted_date = "$date_fields[1] $date_fields[2] $date_fields[3] $date_fields[4]";
	}
	elsif($dateformat eq "MM DD hh:mm:ss") {
	    $formatted_date = "$date_fields[1] $date_fields[2] $date_fields[3]";
	}
	elsif($dateformat eq "MM DD") {
	    $formatted_date = "$date_fields[1] $date_fields[2]";
	}
	else { printError("Invalid Date Format: $dateformat"); }

        $data .=<<EOF;
<TR $virusBackground><TD align=center><input type=checkbox name=qfile value='$key'>
<TD align=center nowrap><input type=image name='viewbutton-$key' src='$GFX_URL/spasm_view.png' alt='View' title='View'>
<input type=image name='deliverbutton-$key' src='$GFX_URL/spasm_deliver.png' alt='Send' title='Send'>
<input type=image name='whiteballbutton-$key' src='$GFX_URL/spasm_whiteball.png' alt='Send+Whitelist' title='Send+Whitelist'>
<input type=image name='deletebutton-$key' src='$GFX_URL/spasm_trash.png' alt='Delete' title='Delete'>
<input type=image name='blackballbutton-$key' src='$GFX_URL/spasm_blackball.png' alt='Delete+Blacklist' title='Delete+Blacklist'>
<TD nowrap>$formatted_date<TD nowrap>$qHash{$key}{sender}<TD nowrap>$qHash{$key}{subject}
EOF
        if($view_sa_score) {
            $data .= "<TD>$qHash{$key}{score}";
        }
	$data .= "</TR>\n";
    }

    if($endentry > $curentry) { $endentry = $curentry; }

    unless($endentry) {
        printHeader();
        printMenuBar();
        print <<EOF;
  $message
  <H1>Quarantine is Empty</H1>
EOF
        printFooter();
        exit;
    }   

    printHeader();
    printMenuBar();
    print <<EOF;
    <P>
    <input type=hidden name=virtdomain value='$virtdomain'>
    <input type=hidden name=source value='viewQuarantine'>
    <input type=hidden name=sort value='$sort'>
    <center>
    $message
    <select name=selectMenu>
      <option value="Send Selected Messages">Send Selected Messages</option>
      <option value="Delete Selected Messages">Delete Selected Messages</option>
      <option value="Send+Whitelist Selected Messages">Send+Whitelist Selected Messages</option>
      <option value="Delete+Blacklist Selected Messages">Delete+Blacklist Selected Messages</option>
      <option value="Delete All Messages">Delete All Messages</option>
    </select>
    <input type=submit name=submit value="Submit">
    </center>
    <p>
EOF
    
    if($endentry) {
        print <<EOF;
<center>Viewing Entries $startentry - $endentry of $totalnum</center>
EOF
    }
    print <<EOF;
    <TABLE border=1 style="background: #ddddff;">
      <TR>
	<TH><input type=checkbox name=selectUnselect title='Check All/Clear All' onClick='javascript:selectAll()'>
	<TH>&nbsp;
	<TH><input type=submit name=submit value="Date" title="Sort by Date" style="width: 100%">
	<TH><input type=submit name=submit value="Sender" title="Sort by Sender" style="width: 100%">
	<TH><input type=submit name=submit value="Subject" title="Sort by Subject" style="width: 100%">
EOF
    if($view_sa_score) { 
        print '<TH><input type=submit name=submit value="SA Score" title="Sort by SpamAssassin Score" style="width: 100%">';
    }
    print <<EOF;
      </TR>
$data
    <tr><td colspan=5 align=left bgcolor='#FFFFFF' style='border:none'><a href="#" onClick='javascript:checkAll(); return false;'>Check All</a> - <a href="#" onClick='javascript:uncheckAll(); return false;'>Clear All</a></tr>
    </table>
    <P>
    <center>
EOF
    if(!$pageall) {  
        if($curentry - ($pageby+1) > 1) {
            print "<input type=submit name=submit value='Previous'>";
        }   
        if($next) {
            print "<input type=submit name=submit value='Next'>";
        }
    }
    print <<EOF;
    <p>
    <select name=selectMenu2>
      <option value="Send Selected Messages">Send Selected Messages</option>
      <option value="Delete Selected Messages">Delete Selected Messages</option>
      <option value="Send+Whitelist Selected Messages">Send+Whitelist Selected Messages</option>
      <option value="Delete+Blacklist Selected Messages">Delete+Blacklist Selected Messages</option>
      <option value="Delete All Messages">Delete All Messages</option>
    </select>
    <input type=submit name=submit2 value="Submit">
    </center>
    <input type=hidden name=entrynum value='$startentry'>
    <input type=hidden name=endentry value='$endentry'>
EOF
    printFooter();
}

# View quarantine file
sub viewQuarantineFile {
    my $resultmsg = shift || '';
    my $qfile = "";
    if($submit eq "View Full Headers" or $submit eq "View Standard Headers" or $resultmsg) {
        $qfile = $cgi->param('qfile')?$cgi->param('qfile'):"";
    } else {
        $qfile = $viewbutton;
    }
    unless($qfile) { printError("No message specified"); }
    my $sender = "";
    my $server = "";
    my $host = "";

    open(IF, "$SPASM_QDIR/$identity/$qfile") or
        printError("Error opening message file $qfile: $!");
    $/ = "\r\n\r\n";
    my $headers = <IF>;
    $headers =~ s/\r//g;
    $headers =~ s/</&lt;/g;
    $headers =~ s/>/&gt;/g;
    $headers =~ s/\n/<br>/g;
    undef $/;
    my $body = <IF>;
    close(IF);

    my $hs = HTML::Strip->new();
    my $bodyclean = $body?$hs->parse($body):"";
    $hs->eof;
    
    $bodyclean =~ s/\r//g;
    $bodyclean =~ s/</&lt;/g;
    $bodyclean =~ s/>/&gt;/g;
    $bodyclean =~ s/\n/<br>/g;

    my $hdrButton;
    if($submit eq "View Full Headers") {
        $hdrButton = "View Standard Headers";
    } else {
        $hdrButton = "View Full Headers";
    }
    
    printHeader();
    printMenuBar();

    print <<EOF;
  <P>
    <FORM action='$SCRIPT' method='POST'>
      <input type=hidden name=username value="$username">
      <input type=hidden name=password value="$password">
      <input type=hidden name=isp value="$isp">
      <input type=hidden name=virtdomain value="$virtdomain">
      <input type=hidden name=qfile value="$qfile">
      $resultmsg
      <TABLE>
        <TR>
          <TD align=center>
            <input type=submit name=submit value="Delete Message">
            <input type=submit name=submit value="Send Message">
            <input type=submit name=submit value="$hdrButton">
        </TR> 
        <TR>
          <TD align=center>
            <input type=submit name=submit value="Blacklist Sender">
            <input type=submit name=submit value="Blacklist Server">
            <input type=submit name=submit value="Blacklist Domain">
        </TR>
        <TR>
          <TD align=center>
            <input type=submit name=submit value="Whitelist Sender">
            <input type=submit name=submit value="Whitelist Server">
            <input type=submit name=submit value="Whitelist Domain">
        </TR>
      </TABLE>
      <P>
EOF
    
    if($submit eq "View Full Headers") {
        print $headers;
        if($headers =~ /X-spasm-sender: (.+?)<br>/) { $sender = $1; }
        if($headers =~ /X-spasm-sender-IP: (.+?)<br>/) { $server = $1; }
        if($headers =~ /X-spasm-sender-host: [^\[].*?([^\.]+\.[^\.]+?)<br>/) { $host = $1; }
    } else {
        my $header1 = "";
        my $header2 = "";
        my $header3 = "";
        my $header4 = "";
        for my $header (split(/<br>/, $headers)) {
            if($header =~ /^From:/i or $header =~ /^To:/i or $header =~ /^Subject:/i or $header =~ /^Date:/i) {
                $header1 .= "$header<br>";
            }
            if($header =~ /^X-spasm-sender:/) {
                $header =~ s/^X-spasm-sender/Sender/;
                print "<br>";
                $header =~ /^Sender: (.*)$/;
                $sender = $1;
                $header2 = "$header<br>";
            }
            if($header =~ /^X-spasm-sender-IP:/) {
                $header =~ s/^X-spasm-sender-IP/Server/;
                $header =~ /^Server: (.*)$/;
                $server = $1;
                $header3 = "$header<br>";
            }
            if($header =~ /^X-spasm-sender-host: [^\[]/) {
                $header =~ s/^X-spasm-sender-host: .*?([^\.]+\.[^\.]+)$/Host: $1/;
                $host = $1;
                $header4 = "$header<br>";
            }
        }
        print "<b>$header2$header3$header4</b><BR><BR>$header1<BR>";
    }
    print $bodyclean;

    print <<EOF;
      <input type=hidden name=sender value="$sender">
      <input type=hidden name=server value="$server">
      <input type=hidden name=host value="$host">
      <input type=hidden name=lastfunction value="viewQuarantineFile">
  </P>
EOF
    printFooter();
}

# Delete all quarantine files
sub clearQuarantine {
    logAction("cleared quarantine");
    opendir(DIR, "$SPASM_QDIR/$identity");
    my @readdir = readdir(DIR);
    closedir(DIR);
    for my $qfile (@readdir) {
        if($qfile =~ /^\./) { next; }
        system($SPASM_CGI_APP, "delete_qfile", "$identity", "$qfile");
    }
    viewQuarantine();
}

# Delete selected quarantine files
sub deleteQuarantineFile {
    my @qfile;
    my $deleted = 0;
    if($deletebutton) { push @qfile, $deletebutton; }
    else { @qfile = $cgi->param('qfile')?$cgi->param('qfile'):(); }
    unless(@qfile) { printError("No message specified"); }
    for my $qfile (@qfile) {
        system($SPASM_CGI_APP, "delete_qfile", "$identity", "$qfile");
	$deleted++;
    }
    logAction("deleted $deleted quarantine files");
    $cgi->param('reload', 1);
    my $startentry = $cgi->param('entrynum')?$cgi->param('entrynum'):1;
    my $endentry = $cgi->param('endentry')?$cgi->param('endentry'):1;
    if($deleted > 0 and $startentry != 1 and ($endentry - $deleted < $startentry)) {
        # We deleted the last entry[s] on the page. Go to previous page.
        $cgi->param('reload', 2);
    }
    viewQuarantine("<H1>Message[s] Deleted</H1>");
}

# Blackball quarantine files
sub blackballQuarantineFile {
    my @qfile;
    my $deleted = 0;
    if($blackballbutton) { push @qfile, $blackballbutton; }
    else { @qfile = $cgi->param('qfile')?$cgi->param('qfile'):(); }
    unless(@qfile) { printError("No message specified"); }
    for my $qfile (@qfile) {
        my $mailfrom;
        open(IF, "<$SPASM_QDIR/$identity/$qfile") or printError("Error opening quarantine file: $!");
        while(<IF>) {
            if(/^X-spasm-sender: (.+)/) { $mailfrom = $1; last; }
        }
        close(IF);
        $mailfrom =~ s/[\r|\n]//g;
        unless($mailfrom) { printError("Unable to find sender"); }
        if($mailfrom ne "<>") { system($SPASM_CGI_APP, "add_blacklist", "$identity", "$mailfrom", "1"); }
        system($SPASM_CGI_APP, "delete_qfile", "$identity", "$qfile");
	$deleted++;
    }
    logAction("blackballed $deleted quarantine files");
    $cgi->param('reload', 1);
    my $startentry = $cgi->param('entrynum')?$cgi->param('entrynum'):1;
    my $endentry = $cgi->param('endentry')?$cgi->param('endentry'):1;
    if($deleted > 0 and $startentry != 1 and ($endentry - $deleted < $startentry)) {
        # We deleted the last entry[s] on the page. Go to previous page.
        $cgi->param('reload', 2);
    }
    viewQuarantine("<H1>Message[s] Deleted and Blacklisted</H1>");
}

# Deliver quarantine files
sub deliverQuarantineFile {
    my @qfile;
    my $deleted = 0;
    if($deliverbutton) { push @qfile, $deliverbutton; }
    else {  @qfile = $cgi->param('qfile')?$cgi->param('qfile'):(); }
    unless(@qfile) { printError("No message specified"); }
    chdir("$SPASM_QDIR/$identity");
    for my $qfile (@qfile) {
        open(IF, "$qfile") or next;
        my $mailfrom = "";
        my $rcptto = "";
        while(<IF>) {
            if(/X-spasm-sender: (\S+)/) {
                $mailfrom = $1;
            }
            if(/X-spasm-sender:/ and !$mailfrom) {
                $mailfrom = "<>";
            }
            if(/X-spasm-rcpt: (\S+)/) {
                $rcptto = $1;
            }
            if($mailfrom and $rcptto) { last; }
        }
        close(IF);
        unless($mailfrom) { printError("Unable to find sender"); }
        unless($rcptto) { printError("Unable to find recipient"); }
        `$SENDMAIL_BIN -f"$mailfrom" -- "$rcptto" < $SPASM_QDIR/$identity/$qfile`;
        system($SPASM_CGI_APP, "delete_qfile", "$identity", "$qfile");
	$deleted++;
    }
    logAction("delivered $deleted quarantine files");
    $cgi->param('reload', 1);
    my $startentry = $cgi->param('entrynum')?$cgi->param('entrynum'):1;
    my $endentry = $cgi->param('endentry')?$cgi->param('endentry'):1;
    if($deleted > 0 and $startentry != 1 and ($endentry - $deleted < $startentry)) {
        # We deleted the last entry[s] on the page. Go to previous page.
        $cgi->param('reload', 2);
    }
    viewQuarantine("<H1>Message[s] Sent</H1>");
}

# Whiteball quarantine files
sub whiteballQuarantineFile {
    my @qfile;
    my $deleted = 0;
    if($whiteballbutton) { push @qfile, $whiteballbutton; }
    else { @qfile = $cgi->param('qfile')?$cgi->param('qfile'):(); }
    unless(@qfile) { printError("No message specified"); }
    for my $qfile (@qfile) {
        my $mailfrom = "";
	my $rcptto = "";
        open(IF, "<$SPASM_QDIR/$identity/$qfile") or printError("Error opening quarantine file: $!");
        while(<IF>) {
            if(/^X-spasm-sender: (\S+)/) { $mailfrom = $1; }
            if(/^X-spasm-sender:/ and !$mailfrom) { $mailfrom = "<>"; }
	    if(/^X-spasm-rcpt: (\S+)/) { $rcptto = $1; }
            if($mailfrom and $rcptto) { last; }
        }
        close(IF);
        unless($mailfrom) { printError("Unable to find sender"); }
	unless($rcptto) { printError("Unable to find recipient"); }
        system($SPASM_CGI_APP, "add_whitelist", "$identity", "$mailfrom", "1");
        `$SENDMAIL_BIN -f"$mailfrom" -- "$rcptto" < $SPASM_QDIR/$identity/$qfile`;
        `/bin/cp $SPASM_QDIR/$identity/$qfile $HAM_DIR/$identity-$qfile`;
        system($SPASM_CGI_APP, "delete_qfile", "$identity", "$qfile");
	$deleted++;
    }
    logAction("whiteballed $deleted quarantine files");
    $cgi->param('reload', 1);
    my $startentry = $cgi->param('entrynum')?$cgi->param('entrynum'):1;
    my $endentry = $cgi->param('endentry')?$cgi->param('endentry'):1;
    if($deleted > 0 and $startentry != 1 and ($endentry - $deleted < $startentry)) {
        # We deleted the last entry[s] on the page. Go to previous page.
        $cgi->param('reload', 2);
    }
    viewQuarantine("<H1>Message[s] Sent and Whitelisted</H1>");
}

# Add entries to blacklist
sub blacklistEntry {
    my $lastfunction = $cgi->param('lastfunction') || '';
    my $count = 0;
    my @logEntries = $cgi->param('logentry')?$cgi->param('logentry'):();
    if(@logEntries) {
	for(@logEntries) {
	    my $entry = `$SPASM_CGI_APP view_log $identity $_`;
	    chomp $entry;
	    my @fields = split(/:/, $entry);
	    if($fields[6] ne "<>") { $count++; system($SPASM_CGI_APP, "add_blacklist", "$identity", "$fields[6]", "1"); }
	}
	logAction("blacklisted $count entries");
	printHeader();
	printMenuBar();
	print "<H1>Senders Blacklisted</H1>";
	printFooter();
    } else {
        my $entry = "";
        my $type = "";
        if($submit eq "Blacklist Sender") {
            my $sender = $cgi->param('sender')?$cgi->param('sender'):"";
            unless($sender) { printError("Error: Sender not found"); }
            $entry = $sender;
            $type = "Sender";
        }
        if($submit eq "Blacklist Domain") {
            my $host = $cgi->param('host')?$cgi->param('host'):"";
            unless($host) { printError("Error: Host not found"); }
            $entry = $host;
            $type = "Host";
        }
        if($submit eq "Blacklist Server") {
            my $server = $cgi->param('server')?$cgi->param('server'):"";
            unless($server) { printError("Error: Server not found"); }
            $entry = $server;
            $type = "Server";
        }
        if($entry ne "<>") { $count++; system($SPASM_CGI_APP, "add_blacklist", "$identity", "$entry", "1"); }
	if($count) { logAction("blacklisted $type $entry"); }
	if($lastfunction eq "viewQuarantineFile") {
	    viewQuarantineFile("<H1>$type $entry Blacklisted</H1>");
	} else {
	    printHeader();
	    printMenuBar();
	    print "<H1>$type $entry Blacklisted</H1>";
	    printFooter();
	}
    }
}

# Add entries to whitelist
sub whitelistEntry {
    my $lastfunction = $cgi->param('lastfunction') || '';
    my $count = 0;
    my @logEntries = $cgi->param('logentry')?$cgi->param('logentry'):();
    if(@logEntries) {
        for(@logEntries) {
            my $entry = `$SPASM_CGI_APP view_log $identity $_`;
            chomp $entry;
            my @fields = split(/:/, $entry);
            system($SPASM_CGI_APP, "add_whitelist", "$identity", "$fields[6]", "1");
	    $count++;
        }
	logAction("whitelisted $count entries");
	printHeader();
	printMenuBar();
	print "<H1>Senders Whitelisted</H1>";
	printFooter();
    } else {
	my $entry = "";
	my $type = "";
	if($submit eq "Whitelist Sender") {
            my $sender = $cgi->param('sender')?$cgi->param('sender'):"";
            unless($sender) { printError("Error: Sender not found"); }
            $entry = $sender;
            $type = "Sender";
        }
        if($submit eq "Whitelist Domain") {
            my $host = $cgi->param('host')?$cgi->param('host'):"";
            unless($host) { printError("Error: Host not found"); }
            $entry = $host;
            $type = "Host";
        }
        if($submit eq "Whitelist Server") {
            my $server = $cgi->param('server')?$cgi->param('server'):"";
            unless($server) { printError("Error: Server not found"); }
            $entry = $server;
            $type = "Server";
        }
        system($SPASM_CGI_APP, "add_whitelist", "$identity", "$entry", "1");
	logAction("whitelisted $type $entry");
	if($lastfunction eq "viewQuarantineFile") {
	    viewQuarantineFile("<H1>$type $entry Whitelisted</H1>");
	} else {
	    printHeader();
	    printMenuBar();
	    print "<H1>$type $entry Whitelisted</H1>";
	    printFooter();
	}
    }
}

# View whitelist
sub viewWhitelist {
    my $message = shift || "";
    my @whitelist = `$SPASM_CGI_APP get_whitelist $identity`;
    my $whitelist = join('', sort sortRoutine @whitelist);
    printHeader();
    printMenuBar();
    print <<EOF;
      <P>
      <input type=hidden name=virtdomain value="$virtdomain">
      <input type=hidden name=virtemail value="$virtemail">
      <input type=hidden name=source value="viewWhitelist">
      <center>                                                                                                        
	<p>$message</p>
        Whitelists are one entry per line, and may be one of the following:
	<table><tr><td>
	<ul>
	  <li>Email Address: 'user\@domain.com'
	  <li>Server Domain: 'domain.com'
	  <li>Email Address Domain: '\@domain.com'
	  <li>Classful Netblock: 10, 10.1, 10.1.2, 10.1.2.3
	  <li>CIDR Address: 10.1.2.0/25
	</ul>
	</tr></table>
	<input type=submit name=submit value="Update"><br>
        <textarea name=whitelist cols=80 rows=20 wrap=off>$whitelist</textarea>
      </center>
EOF
    printFooter();
}

# Update whitelist
sub updateWhitelist {
    my $num = 0;
    my $whitelist = $cgi->param('whitelist')?$cgi->param('whitelist'):"";
    open(PIPE, "|$SPASM_CGI_APP set_whitelist $identity");
    for my $wl (split(/[\n\r]/, $whitelist)) {
        $wl =~ s/[\n\r]//g;
        next unless $wl;
        $wl =~ tr/A-Z/a-z/;
        print PIPE "$wl\n";
	$num++;
    }
    close(PIPE);

    logAction("updated whitelist: $num entries");
    viewWhitelist("<H1>Whitelist Updated</H1>");
}

# View blacklist
sub viewBlacklist {
    my $message = shift || "";
    my @blacklist = `$SPASM_CGI_APP get_blacklist $identity`;
    my $blacklist = join('', sort sortRoutine @blacklist);
    printHeader();
    printMenuBar();
    print <<EOF;
      <P>
      <input type=hidden name=virtdomain value="$virtdomain">
      <input type=hidden name=virtemail value="$virtemail">
      <input type=hidden name=source value="viewBlacklist">
      <center>
	<p>$message</p>
        Blacklists are one entry per line, and may be one of the following:
        <table><tr><td>
        <ul>
          <li>Email Address: 'user\@domain.com'
          <li>Server or Email Address Domain: 'domain.com'
          <li>Classful Netblock: 10, 10.1, 10.1.2, 10.1.2.3
          <li>CIDR Address: 10.1.2.0/25
        </ul>
        </tr></table>
	<input type=submit name=submit value="Update"><br>
        <textarea name=blacklist cols=80 rows=20 wrap=off>$blacklist</textarea>
      </center>
EOF
    printFooter();
}

# Update blacklist
sub updateBlacklist {
    my $num = 0;
    my $blacklist = $cgi->param('blacklist')?$cgi->param('blacklist'):"";
    open(PIPE, "|$SPASM_CGI_APP set_blacklist $identity");
    for my $bl (split(/[\n\r]/, $blacklist)) {
        $bl =~ s/[\n\r]//g;
        next unless $bl;
        $bl =~ tr/A-Z/a-z/;
        print PIPE "$bl\n";
	$num++;
    }
    close(PIPE);

    logAction("updated blacklist: $num entries");
    viewBlacklist("<H1>Blacklist Updated</H1>");
}

# View local blacklist
sub viewLocalBL {
    my @blacklist = `$SPASM_CGI_APP get_blacklist $LOCALBL`;
    my @bl = sort sortRoutine @blacklist;
    printHeader();
    printMenuBar();
    print "<P><center><table>\n";
    for (@bl) { print "<tr><td>$_</tr>\n"; }
    print "</table></center></body></html>\n";
}

# Search global ip blacklist
sub searchGlobalBL {
    my $ipaddress = $cgi->param('ipaddress')?$cgi->param('ipaddress'):"";
    unless($ipaddress) {
	printHeader();
	print <<EOF;
  <body>
    <h1 align=center>Search Global IP Blacklist</h1>
    <p><hr width="75%"></p>
    <form action=$SCRIPT method=GET>
    <p align=center>IP Address: <input type=text size=16 maxlength=15 name=ipaddress></p>
    <p align=center><input type=submit name=submit value="Search Global IP Blacklist"></p>
    </form>
  </body>
</html>
EOF
	exit;
    }
    if($ipaddress !~ /^[\d|\.]+$/) { printMenulessError("Invalid IP Address"); }
    my @ip = split(/\./, $ipaddress);
    my $found = 0;
    for my $i ($ipaddress, "$ip[0].$ip[1].$ip[2]", "$ip[0].$ip[1]", $ip[0]) {
        my $rv = `$SPASM_CGI_APP lookup_blacklist $GLOBAL_IP_BL $i`;
        chomp $rv;
        if($rv) { $found = $i; }
        if($found) { last; }
    }
    printHeader();
    print <<EOF;
  <body>
    <h1 align=center>Search Global IP Blacklist</h1>
    <p><hr width="75%"></p>
EOF
    if($found) {
	print <<EOF;
    <p align=center>Found $ipaddress ($found)</p>
    <form action=$SCRIPT method=POST>
    <input type=hidden name=ipaddress value=$found>
    <input type=hidden name=submit value="deleteGlobalBL">
    <p align=center><input type=submit name=null value="Remove"></p>
    </form>
EOF
    } else {
	print "<p align=center>IP Address Not Found: $ipaddress</p>\n";
    }
    print "</body></html>\n";
}

sub deleteGlobalBL {
    my $ipaddress = $cgi->param('ipaddress')?$cgi->param('ipaddress'):"";
    if($ipaddress !~ /^[\d|\.]+$/) { printMenulessError("Invalid IP Address"); }
    `$SPASM_CGI_APP del_blacklist $GLOBAL_IP_BL $ipaddress`;
    printHeader();
    print <<EOF;
  <body>
    <h1 align=center>Remove Global IP Blacklist</h1>
    <p><hr width="75%"></p>
    <p align=center>Removed $ipaddress</p>
  </body>
</html>
EOF
}

#
# User Preferences
#

sub viewPreferences {
    printHeader();
    printMenuBar();

    if($virtemail) { # Preferences associated with individual addresses
        # Get use_subject_tagging
        my $subject_tagging = `$SPASM_CGI_APP lookup_prefs $identity use_subject_tagging`;
        chomp $subject_tagging;
        my $stag_menu = "<select name=use_subject_tagging>";
        if($subject_tagging) { $stag_menu .= "<option value=off>Off</option><option value=on selected>On</option></select>"; }
        else { $stag_menu .= "<option value=off selected>Off</option><option value=on>On</option></select>"; }

        # Get use_logging
        my $use_logging = `$SPASM_CGI_APP lookup_prefs $identity use_logging`;
        chomp $use_logging;
        my $logging_menu;
        if($use_logging) {
            $logging_menu = "<select name=use_logging><option value='0'>Off</option><option value='1' selected>On</option></select>\n";
        } else {
            $logging_menu = "<select name=use_logging><option value='0' selected>Off</option><option value='1'>On</option></select>\n";
        }

        # Print the page
        print <<EOF;
    <p align=center>Domain Address Preferences</p>
    <table frame=border rules=rows style="background:#ddddff;">
      <tr><th colspan=2 align=center>Preferences</tr>
      <tr><td><a href="#" onClick="openHelpWindow('Subject Tagging', '$SPASM_HELP{use_subject_tagging}'); return false;">Subject Tagging</a><td align=right>$stag_menu</tr>
      <tr><td><a href="#" onClick="openHelpWindow('Logging', '$SPASM_HELP{use_logging}'); return false;">Logging</a><td align=right>$logging_menu</tr>
      </table>
      <p><center><input type=submit name=submit value='Update Preferences'></center></p>
EOF
    } elsif($virtdomain and $virtdomain ne "$username\@$DEFAULT_DOMAIN") { # Preferences associated with domains
        # Get quarantine options
        my $qExpire = `$SPASM_CGI_APP lookup_prefs $identity quarantine_expire`;
        my $reminder = `$SPASM_CGI_APP lookup_prefs $identity email_reminders`;
        chomp $qExpire;
        chomp $reminder;
        unless($qExpire) { $qExpire = 14; }
        unless($reminder) { $reminder = 0; }
        my $qselect;
        for (1..7, 14, 21) {
            if($qExpire == $_) { $qselect .= "<option value=$_ selected>$_ Days</option>"; }
            else { $qselect .= "<option value=$_>$_ Days</option>"; }
        }
        my $rselect;
        for (0..7, 14, 21) {
            if($reminder == $_) { $rselect .= "<option value=$_ selected>$_ Days</option>"; }
            else { $rselect .= "<option value=$_>$_ Days</option>"; }
        }

        # Get log options
        my $lExpire = `$SPASM_CGI_APP lookup_prefs $identity log_expire`;
        chomp $lExpire;
        unless($lExpire) { $lExpire = 14; }
        my $lselect;
        for (1..7, 14, 21) {
            if($lExpire == $_) { $lselect .= "<option value=$_ selected>$_ Days</option>"; }
            else { $lselect .= "<option value=$_>$_ Days</option>"; }
        }

        # Get use_subject_tagging
        my $subject_tagging = `$SPASM_CGI_APP lookup_prefs $identity use_subject_tagging`;
        chomp $subject_tagging;
        my $stag_menu = "<select name=use_subject_tagging>";
        if($subject_tagging) { $stag_menu .= "<option value=off>Off</option><option value=on selected>On</option></select>"; }
        else { $stag_menu .= "<option value=off selected>Off</option><option value=on>On</option></select>"; }

        # Get use_logging
        my $use_logging = `$SPASM_CGI_APP lookup_prefs $identity use_logging`;
        chomp $use_logging;
        my $logging_menu;
        if($use_logging) {
            $logging_menu = "<select name=use_logging><option value='0'>Off</option><option value='1' selected>On</option></select>\n";
        } else {
            $logging_menu = "<select name=use_logging><option value='0' selected>Off</option><option value='1'>On</option></select>\n";
        }

        # Get enable_virtual_addresses
        my $enable_virtual_addresses = `$SPASM_CGI_APP lookup_prefs $identity enable_virtual_addresses`;
        chomp $enable_virtual_addresses;
        my $enable_virtual_addresses_menu;
        if($enable_virtual_addresses) {
            $enable_virtual_addresses_menu = "<select name=enable_virtual_addresses><option value='0'>Off</option><option value='1' selected>On</option></select>\n";
        } else {
            $enable_virtual_addresses_menu .= "<select name=enable_virtual_addresses><option value='0' selected>Off</option><option value='1'>On</option></select>\n";
        }

        # Print the page
        print <<EOF;
    <p align=center>Domain Preferences</p>
    <table frame=border rules=rows style="background:#ddddff;">
      <tr><th colspan=2 align=center>Preferences</tr>
      <tr><td><a href="#" onClick="openHelpWindow('Subject Tagging', '$SPASM_HELP{use_subject_tagging}'); return false;">Subject Tagging</a><td align=right>$stag_menu</tr>
      <tr><td><a href="#" onClick="openHelpWindow('Logging', '$SPASM_HELP{use_logging}'); return false;">Logging</a><td align=right>$logging_menu</tr>
      <tr><td>Enable Virtual Addresses<td align=right>$enable_virtual_addresses_menu</tr>
      <tr style="background:#ffffff;"><td colspan=2>&nbsp;</tr>
      <tr><th colspan=2 align=center>Quarantine Options</tr>
      <tr><td>Quarantine Messages Expire In:<td align=right><select name=qexpire>$qselect</select></tr>
      <tr><td>Quarantine Reminders Sent Every:<td align=right><select name=reminder>$rselect</select></tr>
      <tr><td>Log Entries Expire In:<td align=right><select name=lexpire>$lselect</select></tr>
      <tr><td>Clear All Quarantine Messages:<td align=right><input type=submit name=submit value="Delete All Messages"></tr>
      </table>
      <p><center><input type=submit name=submit value='Update Preferences'></center></p>
EOF

    } else { # Preferences associated with login

        # Get default domain account
        my $defaultdomainaccount = `$SPASM_CGI_APP lookup_prefs $username\@$DEFAULT_DOMAIN defaultdomainaccount`;
        chomp $defaultdomainaccount;
        unless($defaultdomainaccount) { $defaultdomainaccount = "$username\@$DEFAULT_DOMAIN"; }
        my $domainoptions = "";
        if(@domains) {
            $domainoptions = "<select name='defaultdomainaccount'>";
            for (@domains) {
                if($_ eq $defaultdomainaccount) { $domainoptions .= "<option selected value='$_'>$_"; }
                else { $domainoptions .= "<option value='$_'>$_"; }
            }
            $domainoptions .= "</select>";
        }

        # Get aliases
        my $aliases = `$SPASM_CGI_APP lookup_prefs $username\@$DEFAULT_DOMAIN aliases`;
        chomp $aliases; 
        my @aliases = split(/,/, $aliases);

        # Get quarantine options
        my $qExpire = `$SPASM_CGI_APP lookup_prefs $identity quarantine_expire`;
        my $reminder = `$SPASM_CGI_APP lookup_prefs $identity email_reminders`;
        chomp $qExpire;
        chomp $reminder;
        unless($qExpire) { $qExpire = 14; }
        unless($reminder) { $reminder = 0; }
        my $qselect;
        for (1..7, 14, 21) {
            if($qExpire == $_) { $qselect .= "<option value=$_ selected>$_ Days</option>"; }
            else { $qselect .= "<option value=$_>$_ Days</option>"; }
        }
        my $rselect;
        for (0..7, 14, 21) {
            if($reminder == $_) { $rselect .= "<option value=$_ selected>$_ Days</option>"; }
            else { $rselect .= "<option value=$_>$_ Days</option>"; }
        }

        # Get log options
        my $lExpire = `$SPASM_CGI_APP lookup_prefs $identity log_expire`;
        chomp $lExpire;
        unless($lExpire) { $lExpire = 14; }
        my $lselect;
        for (1..7, 14, 21) {
            if($lExpire == $_) { $lselect .= "<option value=$_ selected>$_ Days</option>"; }
            else { $lselect .= "<option value=$_>$_ Days</option>"; }
        }

        # Get use_subject_tagging
        my $subject_tagging = `$SPASM_CGI_APP lookup_prefs $identity use_subject_tagging`;
        chomp $subject_tagging;
        my $stag_menu = "<select name=use_subject_tagging>";
        if($subject_tagging) { $stag_menu .= "<option value=off>Off</option><option value=on selected>On</option></select>"; }
        else { $stag_menu .= "<option value=off selected>Off</option><option value=on>On</option></select>"; }

        # Get use_logging
        my $use_logging = `$SPASM_CGI_APP lookup_prefs $identity use_logging`;
        chomp $use_logging;
        my $logging_menu; 
        if($use_logging) {
            $logging_menu = "<select name=use_logging><option value='0'>Off</option><option value='1' selected>On</option></select>\n";
        } else {
            $logging_menu = "<select name=use_logging><option value='0' selected>Off</option><option value='1'>On</option></select>\n";
        }

        # Get view_sa_score
        my $view_sa_score = `$SPASM_CGI_APP lookup_prefs $identity view_sa_score`;
        chomp $view_sa_score;
        my $sa_score_menu;
        if($view_sa_score) {
            $sa_score_menu = "<select name=view_sa_score><option value='0'>Off</option><option value='1' selected>On</option></select>\n";
        } else {
            $sa_score_menu = "<select name=view_sa_score><option value='0' selected>Off</option><option value='1'>On</option></select>\n";
        }

        # Get cookie options
        my $setcookie = `$SPASM_CGI_APP lookup_prefs $username\@$DEFAULT_DOMAIN setcookie`; 
        chomp $setcookie;
        if($setcookie eq "yes") { $setcookie = "checked"; }
        else { $setcookie = ""; }

        # Get pageby options
        my $pageby = `$SPASM_CGI_APP lookup_prefs $username\@$DEFAULT_DOMAIN pageby`;
        chomp $pageby;
        unless($pageby) { $pageby = 25; }
        my $pagebymenu = "<select name=pageby>";
        if($pageby eq 'All') {
            $pagebymenu .= "<option value=All selected>All</option>";
            for(25,50,100,250) {
                $pagebymenu .= "<option value=$_>$_</option>";
            }
        } else {
            for (25, 50, 100, 250, "All") {
                if($_ eq "All") {
                    $pagebymenu .= "<option value=$_";
                    $pagebymenu .= ($pageby eq $_?" selected":"");
                    $pagebymenu .= ">All</option>";
                } else {
                    $pagebymenu .= "<option value=$_";
                    $pagebymenu .= ($pageby==$_?" selected":"");
                    $pagebymenu .= ">$_</option>";
                }
            }
        }
        $pagebymenu .= "</select>";

        # Get filtersort options
        my $filtersort = `$SPASM_CGI_APP lookup_prefs $username\@$DEFAULT_DOMAIN filtersort`;
        chomp $filtersort;
        unless($filtersort) { $filtersort = "Default"; }
        my $filtersortmenu = "<select name=filtersort>";
        for ("Default", "Alphabetical") {
            $filtersortmenu .= "<option value=$_";
            $filtersortmenu .= ($filtersort eq $_?" selected":"");
            $filtersortmenu .= ">$_</option>";
        }

        # Fonts
        my @f_sizes = ("8pt", "9pt", "10pt", "11pt", "12pt", "13pt", "14pt", "15pt", "16pt");

        my $h1_font = `$SPASM_CGI_APP lookup_prefs $username\@$DEFAULT_DOMAIN h1_font`;
        chomp $h1_font;
        unless($h1_font) { $h1_font = "13pt"; }
        my $h1_font_menu = "<select name=h1_font>";
        for (@f_sizes) {
            $h1_font_menu .= "<option value=$_";
            $h1_font_menu .= ($h1_font eq $_?" selected":"");
            $h1_font_menu .= ">$_</option>";
        }

        my $b_font = `$SPASM_CGI_APP lookup_prefs $username\@$DEFAULT_DOMAIN b_font`;
        chomp $b_font;
        unless($b_font) { $b_font = "12pt"; }
        my $b_font_menu = "<select name=b_font>";
        for (@f_sizes) {
            $b_font_menu .= "<option value=$_";
            $b_font_menu .= ($b_font eq $_?" selected":"");
            $b_font_menu .= ">$_</option>";
        }

        # Date Format
        my $dateformat = `$SPASM_CGI_APP lookup_prefs $username\@$DEFAULT_DOMAIN dateformat`;
        chomp $dateformat;
        unless($dateformat) { $dateformat = "Day MMM DD hh:mm:ss YYYY"; }
        my $dateformat_menu = "<select name=dateformat>";
        for (@DATE_FORMATS) {
            $dateformat_menu .= "<option value='$_'";
            $dateformat_menu .= ($dateformat eq $_?" selected":"");
            $dateformat_menu .= ">$_</option>";
        }

        # Maximum Subject Length
        my $max_subject_length = `$SPASM_CGI_APP lookup_prefs $username\@$DEFAULT_DOMAIN max_subject_length`;
        chomp $max_subject_length;
        unless($max_subject_length) { $max_subject_length = "0"; }

        # Maximum Sender Length
        my $max_sender_length = `$SPASM_CGI_APP lookup_prefs $username\@$DEFAULT_DOMAIN max_sender_length`;
        chomp $max_sender_length;
        unless($max_sender_length) { $max_sender_length = "0"; }

        # Print the page
        print <<EOF;
    <p align=center>Login Preferences</p>
    <table frame=border rules=rows style="background:#ddddff;">
      <tr><th colspan=2 align=center>Preferences</tr>
      <tr><td>Keep me logged in (uses cookies)<td align=right><input type=checkbox name=setcookie value='yes' $setcookie></tr>
      <tr><td>Log/Quarantine entries per page:<td align=right>$pagebymenu</tr>
EOF
        if($domainoptions) { print "<tr><td>Default Domain Account<td align=right>$domainoptions</tr>\n"; }
        print <<EOF;
      <tr><td>Date Format<td align=right>$dateformat_menu</tr>
      <tr><td>Maximum Subject Length (0 for full subject)<td align=right><input type=text size=3 maxlength=2 name=max_subject_length value='$max_subject_length'></tr>
      <tr><td>Maximum Sender Length (0 for full sender)<td align=right><input type=text size=3 maxlength=2 name=max_sender_length value='$max_sender_length'></tr>
      <tr><td><a href="#" onClick="openHelpWindow('Subject Tagging', '$SPASM_HELP{use_subject_tagging}'); return false;">Subject Tagging</a><td align=right>$stag_menu</tr>
      <tr><td><a href="#" onClick="openHelpWindow('Logging', '$SPASM_HELP{use_logging}'); return false;">Logging</a><td align=right>$logging_menu</tr>
      <tr><td>SpamAssassin Score in Quarantine<td align=right>$sa_score_menu</tr>
      <tr><td>Mail Filters Sort Method<td align=right>$filtersortmenu</tr>
      <tr style="background:#ffffff;"><td colspan=2>&nbsp;</tr>
      <tr><th colspan=2 align=center>Fonts</tr>
      <tr><td>Headers<td align=right>$h1_font_menu</tr>
      <tr><td>Body Text<td align=right>$b_font_menu</tr>
      <tr style="background:#ffffff;"><td colspan=2>&nbsp;</tr>
      <tr><th colspan=2 align=center>Quarantine Options</tr>
      <tr><td>Quarantine Messages Expire In:<td align=right><select name=qexpire>$qselect</select></tr>
      <tr><td>Quarantine Reminders Sent Every:<td align=right><select name=reminder>$rselect</select></tr>
      <tr><td>Log Entries Expire In:<td align=right><select name=lexpire>$lselect</select></tr>
      <tr><td>Clear All Quarantine Messages:<td align=right><input type=submit name=submit value="Delete All Messages"></tr>
      <tr style="background:#ffffff;"><td colspan=2>&nbsp;</tr>
      <tr><th colspan=2 align=center>Aliased Accounts <a href="#" onClick="openHelpWindow('Aliased Accounts', '$SPASM_HELP{ALIASED_ACCOUNTS}'); return false;">?</a></tr>
EOF

        if(@aliases) {
            for my $alias (@aliases) {
                print "<tr><td>$alias<td align=right><input type=image name='delalias-$alias' src='$GFX_URL/spasm_trash.png' alt='Delete' title='Delete'\n";
            }
        } else {
            print "<tr><td colspan=2 align=center>[No Aliased Accounts]</tr>\n";
        }
        print "<tr><td colspan=2 align=center><input type=submit name=submit value='Add Aliased Account'></tr>\n";
        print "</table>\n";
        print "<p><center><input type=submit name=submit value='Update Preferences'></center></p>\n";
    }

    printFooter();
}

sub updatePreferences {
    # subject tagging
    my $use_subject_tagging = $cgi->param('use_subject_tagging')?$cgi->param('use_subject_tagging'):"";
    if($use_subject_tagging eq "off") { `$SPASM_CGI_APP set_prefs $identity use_subject_tagging "0"`; }
    elsif($use_subject_tagging eq "on") { `$SPASM_CGI_APP set_prefs $identity use_subject_tagging "1"`; }

    # logging
    my $use_logging = $cgi->param('use_logging');
    if(defined $use_logging) { `$SPASM_CGI_APP set_prefs $identity use_logging "$use_logging"`; }

    if(!$virtemail) {
        # Log options
        my $lExpire = $cgi->param('lexpire')?$cgi->param('lexpire'):0;
        unless($lExpire) { printError("Invalid Log Expire Time"); }
        unless(($lExpire > 0 and $lExpire < 8) or $lExpire == 14 or $lExpire == 21) {
            printError("Invalid Log Expire Time");
        }
        `$SPASM_CGI_APP set_prefs $identity log_expire $lExpire`;

        # Quarantine options
        my $qExpire = $cgi->param('qexpire')?$cgi->param('qexpire'):0;
        unless($qExpire) { printError("Invalid Quarantine Expire Time"); }
        unless(($qExpire > 0 and $qExpire < 8) or $qExpire == 14 or $qExpire == 21) {
            printError("Invalid Quarantine Expire Time");
        }
        `$SPASM_CGI_APP set_prefs $identity quarantine_expire $qExpire`;

        my $reminder = $cgi->param('reminder')?$cgi->param('reminder'):0;
        unless($reminder =~ /^\d+$/) {
            printError("Invalid Reminder Time");
        }
        `$SPASM_CGI_APP set_prefs $identity email_reminders $reminder`;

        # Virtual addresses option
        my $enable_virtual_addresses = $cgi->param('enable_virtual_addresses');
        if(defined $enable_virtual_addresses) { `$SPASM_CGI_APP set_prefs $identity enable_virtual_addresses $enable_virtual_addresses`; }
    }

    if($identity eq "$username\@$DEFAULT_DOMAIN") {
        # Cookies
        my $setcookie = $cgi->param('setcookie')?$cgi->param('setcookie'):"no";
        unless($setcookie eq "yes") { $setcookie = "no"; }
        `$SPASM_CGI_APP set_prefs $identity setcookie $setcookie`;
        if($setcookie eq "yes") {
            $COOKIE_ID = $cgi->cookie(-name=>'spasmLoginID', -value=>"$username", -expires=>'+7d', -path=>"$SCRIPT");
            $COOKIE_PASS = $cgi->cookie(-name=>'spasmLoginPass', -value=>"$password", -expires=>'+7d', -path=>"$SCRIPT");
            if(%ISPS) {
                $COOKIE_ISP = $cgi->cookie(-name=>'spasmLoginISP', -value=>"$isp", -expires=>'+7d', -path=>"$SCRIPT");
            }
        } elsif($COOKIE_ID or $COOKIE_PASS or $COOKIE_ISP) {
            $COOKIE_ID = $cgi->cookie(-name=>'spasmLoginID', -value=>"", -expires=>'0', -path=>"$SCRIPT");
            $COOKIE_PASS = $cgi->cookie(-name=>'spasmLoginPass', -value=>"", -expires=>'0', -path=>"$SCRIPT");
            $COOKIE_ISP = $cgi->cookie(-name=>'spasmLoginISP', -value=>"", -expires=>'0', -path=>"$SCRIPT");
        }

        # pageby
        my $pageby = $cgi->param('pageby')?$cgi->param('pageby'):25;
        `$SPASM_CGI_APP set_prefs $identity pageby $pageby`;

        # default domain account
        my $defaultdomainaccount = $cgi->param('defaultdomainaccount')?$cgi->param('defaultdomainaccount'):"";
        if($defaultdomainaccount) { `$SPASM_CGI_APP set_prefs $identity defaultdomainaccount $defaultdomainaccount`; }

        # date format
        my $dateformat = $cgi->param('dateformat')?$cgi->param('dateformat'):"";
        if($dateformat) { `$SPASM_CGI_APP set_prefs $identity dateformat "$dateformat"`; }

        # max subject length
        my $max_subject_length = $cgi->param('max_subject_length')?$cgi->param('max_subject_length'):"0";
        if($max_subject_length !~ /^\d+$/ or $max_subject_length < 0) { printError("Subject length must be a number greater than or equal to zero"); }
        `$SPASM_CGI_APP set_prefs $identity max_subject_length "$max_subject_length"`;

        # max sender length
        my $max_sender_length = $cgi->param('max_sender_length')?$cgi->param('max_sender_length'):"0";
        if($max_sender_length !~ /^\d+$/ or $max_sender_length < 0) { printError("Sender length must be a number greater than or equal to zero"); }
        `$SPASM_CGI_APP set_prefs $identity max_sender_length "$max_sender_length"`;

	# view_sa_score
	my $view_sa_score = $cgi->param('view_sa_score');
	if(defined $view_sa_score) { `$SPASM_CGI_APP set_prefs $identity view_sa_score "$view_sa_score"`; }

        # filtersort
        my $filtersort = $cgi->param('filtersort')?$cgi->param('filtersort'):"Default";
        `$SPASM_CGI_APP set_prefs $identity filtersort "$filtersort"`;

        # Fonts
        my $h1_font = $cgi->param('h1_font')?$cgi->param('h1_font'):"13pt";
        my $b_font = $cgi->param('b_font')?$cgi->param('b_font'):"12pt";
        `$SPASM_CGI_APP set_prefs $username\@$DEFAULT_DOMAIN h1_font $h1_font`;
        `$SPASM_CGI_APP set_prefs $username\@$DEFAULT_DOMAIN b_font $b_font`;
        $CSS = <<EOF;
      BODY { background: white; color: black; font-family: "times new roman", serif; font-size: $b_font; }
      H1 { font-size: $h1_font; text-align: center; }
      H2 { font-size: 10pt; text-align: center; }
      IMG { border: none; }
      TABLE { margin-left: auto; margin-right: auto; }
      FORM { margin-bottom: 0; }
      A:link { text-decoration: none; color: #004080; }
      A:visited { text-decoration: none; color: #004080; }
      A:hover { text-decoration: none; color: #990000; background: #ddddff; }
      A:active { text-decoration: none; color: #004080; }
EOF
    }

    viewPreferences();
}

sub delPrefsAlias {
    `$SPASM_CGI_APP del_prefs $delaliasbutton alias`;
    my $aliases = `$SPASM_CGI_APP lookup_prefs $username\@$DEFAULT_DOMAIN aliases`;
    chomp $aliases;
    my @aliases = split(/,/, $aliases);
    my @newaliases = ""; 
    for my $a (@aliases) { 
        if($a ne $delaliasbutton) { push @newaliases, $a; }
    }   
    my $newaliases = join(',', @newaliases);
    `$SPASM_CGI_APP set_prefs $username\@$DEFAULT_DOMAIN aliases "$newaliases"`;
    viewPreferences();
}

sub addPrefsAlias {
    my $aliasusername = $cgi->param('aliasusername')?$cgi->param('aliasusername'):"";
    my $aliaspassword = $cgi->param('aliaspassword')?$cgi->param('aliaspassword'):"";
    my $aliasisp = $cgi->param('aliasisp')?$cgi->param('aliasisp'):"";
    my $authorise = 0;

    $aliasusername = lc($aliasusername);

    if(%ISPS) { if($aliasusername and $aliaspassword and $aliasisp) { $authorise = 1; } }
    else { if($aliasusername and $aliaspassword) { $authorise = 1; } }

    if($authorise) {
	my $auth = 0;
	if(%ISPS) {
	    $auth = verifyAccount($aliasusername, $aliaspassword, $aliasisp);
	} else {
	    $auth = verifyAccount($aliasusername, $aliaspassword);
	}
	unless($auth) { printError("Authorisation Failed"); }
	my $aliasdomain = %ISPS?$ISPS{$aliasisp}{domain}:$DEFAULT_DOMAIN;
	if("$aliasusername\@$aliasdomain" eq $identity or "$aliasusername\@$aliasdomain" eq "$username\@$DEFAULT_DOMAIN") {
	    printError("Cannot add account to itself");
	}
	`$SPASM_CGI_APP set_prefs $aliasusername\@$aliasdomain alias $identity`;
        my $aliases = `$SPASM_CGI_APP lookup_prefs $username\@$DEFAULT_DOMAIN aliases`;
        chomp $aliases;
        my @aliases = split(/,/, $aliases);
        push @aliases, "$aliasusername\@$aliasdomain";
        my $newaliases = join(',', @aliases);
        `$SPASM_CGI_APP set_prefs $username\@$DEFAULT_DOMAIN aliases "$newaliases"`;
	viewPreferences();
    } else {
	printHeader();
	printMenuBar();

	my $selectbox = "";
	if(%ISPS) {
	    $selectbox = "<select name=aliasisp>\n";
	    for my $key (sort { $ISPS{$a}{order} <=> $ISPS{$b}{order} } keys(%ISPS)) {
		if($key eq $isp) { $selectbox .= "<option value='$key' selected>$key</option>"; }
		else { $selectbox .= "<option value='$key'>$key</option>"; }
	    }
	    $selectbox .= "</select>\n";
	}

	print <<EOF;
    <table>
    <tr><td>Username:<td><input type=text size=16 name=aliasusername></tr>
    <tr><td>Password:<td><input type=password size=16 name=aliaspassword></tr>
EOF
	if($selectbox) { print "<tr><td>ISP:<td>$selectbox</tr>\n"; }
	print <<EOF;
    </table>
    <p><center><input type=submit name=submit value='Add Aliased Account'></center></p>
EOF

	printFooter();
    }
}

#
# The Virtual Domain functions:
#

# Verify that the domain portion of an email address matches the authenticated domain
sub verifyEmail {
    my $address = shift;
    my ($a, $b) = split(/\@/, $address);
    if($b eq $virtdomain) { return(1); }
    else { return(0); }
}

# View virtual domain email addresses
sub viewEmailAliases {
    if($virtdomain =~ /@/) { printError("'$virtdomain' is not a valid domain!"); }
    unless($virtemail) { $virtemail = $cgi->param('emailAlias')?$cgi->param('emailAlias'):""; }
    my $catchall = "";
    printHeader();
    printMenuBar();
    print <<EOF;
    </FORM>
    <FORM action='$SCRIPT' method='POST'>
    <input type=hidden name=username value="$username">
    <input type=hidden name=password value="$password">
    <input type=hidden name=isp value="$isp">
    <input type=hidden name=virtdomain value="$virtdomain">

    <TABLE frame=border rules=rows style="background:#ddddff">
      <TR><TH>&nbsp;<TH>Virtual Domain Address<TH>Delivered To</TR>
EOF
    
    for my $key (sort(keys %addresses)) {
	my $virt = $key;
	my $real = $addresses{$key};
        if($virt =~ /^\@/) { $catchall = $real; next; }
        print "<TR><TD align=center>";
        if($virt eq $virtemail) { print "<input type=radio name=virtemail value='$virt' checked>"; }
        else { print "<input type=radio name=virtemail value='$virt'>"; }
        print "<TD align=center>$virt<TD align=center>$real</TR>";
    }
    
    unless($catchall) { $catchall = "[none]"; }

    print <<EOF;
    </TABLE>
    <center>
    <input type=submit name=submit value="Edit">
    <input type=submit name=submit value="Delete">
    </center>
    </FORM>
    <P>
    <TABLE>
      <TR><TD align=center colspan=2>Current Default Alias: $catchall</TR>
    </TABLE>
    <P>
    <TABLE>
      <TR><TD align=center><FORM action='$SCRIPT' method='POST'>
                <input type=hidden name=username value="$username">
                <input type=hidden name=isp value="$isp">
                <input type=hidden name=password value="$password">
                <input type=hidden name=virtdomain value="$virtdomain">
                <input type=hidden name=virtemail value="\@$virtdomain">
                <input type=hidden name=realEmail value="$catchall">
                <input type=submit name=submit value="Change Default Alias">
	  </FORM>
          <TD align=center><FORM action='$SCRIPT' method='POST'>
                <input type=hidden name=username value="$username">
                <input type=hidden name=isp value="$isp">
                <input type=hidden name=password value="$password">
                <input type=hidden name=virtdomain value="$virtdomain">
                <input type=submit name=submit value="Add New Alias">
	  </FORM>
      </TR>
    </TABLE>
EOF
    printFooter();
}

# Add a virtual email address
sub addEmailAlias {
    if($virtdomain =~ /@/) { printError("'$virtdomain' is not a valid domain!"); }
    my $addVirtEmail = $cgi->param('addVirtEmail')?$cgi->param('addVirtEmail'):"";
    my $addRealEmail = $cgi->param('addRealEmail')?$cgi->param('addRealEmail'):"";
    $addVirtEmail =~ s/\s//g;
    if($addRealEmail =~ /^\s+$/) { $addRealEmail = ""; }

    unless($addVirtEmail and $addRealEmail) {
        printHeader();
        printMenuBar();
        print <<EOF;
    </FORM>
    <FORM action='$SCRIPT' method='POST'>
      <input type=hidden name=username value="$username">
      <input type=hidden name=isp value="$isp">
      <input type=hidden name=password value="$password">
      <input type=hidden name=virtdomain value="$virtdomain">
      <TABLE>                                                                                                       
        <TR><TD align=center>Virtual Domain Address:<TD><input type=text size=30 maxlength=100 name=addVirtEmail value="$addVirtEmail"></TR>
        <TR><TD>&nbsp<TD>(e.g. "username\@$virtdomain")</TR>
        <TR><TD>&nbsp<TD>&nbsp</TR>
        <TR><TD align=center>Deliver To:<TD><input type=text size=30 maxlength=100 name=addRealEmail value="$addRealEmail"></TR>
        <TR><TD>&nbsp<TD>(e.g. "username\@$DEFAULT_DOMAIN")</TR>
      </TABLE>
      <TABLE>
        <TR><TD><input type=submit name=submit value="Save Alias"></TR>
      </TABLE>
EOF
        printFooter();
        exit;
    }
    unless($addVirtEmail =~ /\@/) { $addVirtEmail .= "\@$virtdomain"; }
    unless(verifyEmail($addVirtEmail)) { printError("Invalid Virtual Domain Address"); }
    my $ldap = Net::LDAP->new($LDAP_SERVER, port=>$LDAP_PORT) or printError("Error connecting to LDAP server: $@");
    my $result = $ldap->bind($LDAP_BINDDN, password=>$LDAP_BINDPW);
    if($result->code) { printError("Error binding to LDAP server: ", $result->error); }
    $result = $ldap->add(dn=>"virtEmail=$addVirtEmail,$LDAP_VIRTDN",
                            attr=>[ 'objectclass' => 'virtUserTable',
                                    'realEmail' => "$addRealEmail",
                                    'virtEmail' => "$addVirtEmail" ] );
    if($result->code) { printError("Error adding mailbox: ", $result->error); }

    logAction("added mailbox $addVirtEmail -> $addRealEmail");

    printHeader();                                                                                                  
    printMenuBar();
    print "<H1>Alias Added</H1>";
    printFooter();
}

# Edit a virtual domain email address
sub editEmailAlias {
    if($virtdomain =~ /@/) { printError("'$virtdomain' is not a valid domain!"); }
    unless($virtemail) { printError("Virtual domain email address not found"); }
    unless(verifyEmail($virtemail)) { printError("You do not have permission to edit that address"); }
    my $newRealEmail = $cgi->param('newRealEmail')?$cgi->param('newRealEmail'):"";
    unless($newRealEmail) {
        printHeader();
        printMenuBar();
        print <<EOF;
    </FORM>
    <FORM action='$SCRIPT' method='POST'>
      <input type=hidden name=username value="$username">
      <input type=hidden name=isp value="$isp">
      <input type=hidden name=password value="$password">
      <input type=hidden name=virtdomain value="$virtdomain">
      <input type=hidden name=virtemail value="$virtemail">
      <TABLE>
        <TR><TD align=center>Virtual Domain Address:<TD align=center>$virtemail</TR>
        <TR><TD align=center>New Destination:<TD><input type=text name=newRealEmail></TR>
      </TABLE>
      <P>
      <TABLE>
        <TR><TD><input type=submit name=submit value="Edit"></TR>
      </TABLE>
EOF
        printFooter();
        exit;
    }
    my %changes;
    $changes{realEmail} = $newRealEmail;
    my $ldap = Net::LDAP->new($LDAP_SERVER, port=>$LDAP_PORT) or printError("Error connecting to LDAP server: $@");
    my $result = $ldap->bind($LDAP_BINDDN, password=>$LDAP_BINDPW);
    if($result->code) { printError("Error binding to LDAP server: ", $result->error); }

    $result = $ldap->modify(dn=>"virtEmail=$virtemail,$LDAP_VIRTDN",
                            replace=>{%changes});
    if($result->code) { printError("Error saving changes: ", $result->error); }
    logAction("edited mailbox $virtemail -> $newRealEmail");
    printHeader();
    printMenuBar();
    print "<H1>Address $virtemail Changed</H1>";
    printFooter();
}

# Delete a virtual domain email address
sub delEmailAlias {
    if($virtdomain =~ /@/) { printError("'$virtdomain' is not a valid domain!"); }
    unless($virtemail) { printError("Virtual domain email address not found"); }
    unless(verifyEmail($virtemail)) { printError("You do not have permission to delete that address"); }
    my $ldap = Net::LDAP->new($LDAP_SERVER, port=>$LDAP_PORT) or printError("Error connecting to LDAP server: $@");
    my $result = $ldap->bind($LDAP_BINDDN, password=>$LDAP_BINDPW);
    if($result->code) { printError("Error binding to LDAP server: ", $result->error); }
    $result = $ldap->delete(dn=>"virtEmail=$virtemail,$LDAP_VIRTDN");
    if($result->code) { printError("Error deleting mailbox: ", $result->error); }
    logAction("deleted mailbox $virtemail");
    printHeader();
    printMenuBar();
    print "<H1>Address $virtemail Deleted</H1>";
    printFooter();
}
