#include <dirent.h>
#include <errno.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

#include "spasm.h"

struct spasmConfigStruct *spasm_config;

void
spasm_error(int syslevel, char *errstr, ...)
{   
    va_list arglist;

    va_start(arglist, errstr);
    vfprintf(stderr, errstr, arglist);
    fprintf(stderr, "\n");      /* spasm_error args aren't terminated by \n */
    va_end(arglist);
}

int
main(int argc, char **argv)
{
    DIR *ldir;
    char ldirname[FILENAME_MAX+1];
    struct dirent *dent;
    time_t curtime;
    int rv = 0;

    if((spasm_config = spasm_cfg()) == NULL) {
	exit(1);
    }

    /* Try not to bog down the system */
    nice(5);

    curtime = time(NULL);

    rv = snprintf(ldirname, sizeof(ldirname), "%s/users", spasm_config->spasmroot);
    if(rv < 0) fprintf(stderr, "Warning: truncated ldirname\n");
    if((ldir = opendir(ldirname)) == NULL) {
	fprintf(stderr, "Error opening log directory: %m\n");
	exit(1);
    }

    /* Loop through log directory */
    while((dent = readdir(ldir)) != NULL) {
	char dbfile[FILENAME_MAX+1];
	struct stat filestat;
	struct spasm_list *first, *cur;
	int lexpire;
	char *buf;

	if(strcmp(dent->d_name, ".") == 0 || strcmp(dent->d_name, "..") == 0) continue;

	rv = snprintf(dbfile, sizeof(dbfile), "%s/%s/logs", ldirname, dent->d_name);
        if(rv < 0) fprintf(stderr, "Warning: truncated dbfile\n");
	if(stat(dbfile, &filestat) < 0) {
	    if(errno != ENOENT) fprintf(stderr, "Error getting file status (%s): %m\n", dbfile);
	    continue;
	}
	/* Get log expiration time (in days) */
	if((buf = spasm_lookup_prefs(dent->d_name, "log_expire")) == NULL) {
	    lexpire = spasm_config->defaultlexpire;
	} else {
	    lexpire = atoi(buf);
	    free(buf);
	}
	if(lexpire < 1) lexpire = spasm_config->defaultlexpire;
	/* Convert to seconds */
	lexpire *= 86400;
	/* Loop through user's log */
	cur = qdbm_dump_entries(dbfile, 1);
	while(cur != NULL) {
	    unsigned long stamp;
	    char chr = cur->entry[10];
	    cur->entry[10] = '\0';
	    stamp = strtoul(cur->entry, (char**)NULL, 0);
	    cur->entry[10] = chr;
	    if(stamp < (curtime - lexpire)) qdbm_delete_entry(dbfile, cur->entry);
	    first = cur->next;
	    free(cur);
	    cur = first;
	}
    }
    closedir(ldir);

    exit(0);
}
