#include <arpa/inet.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <sysexits.h>
#include <syslog.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <unistd.h>

#include "spasm.h"
#include "getopt-portable.h"

static char *progname;
struct spasmConfigStruct *spasm_config;

void
spasm_error(int syslevel, char *errstr, ...)
{   
    va_list arglist;

    va_start(arglist, errstr);
    vfprintf(stderr, errstr, arglist);
    fprintf(stderr, "\n");      /* spasm_error args aren't terminated by \n */
    va_end(arglist);
}

void
add_blacklist(char *blacklist, char *entry, char *val)
{
    spasm_insert_bl(blacklist, entry, val);
}

void
del_blacklist(char *blacklist, char *entry)
{
    spasm_delete_bl(blacklist, entry);
}

void
lookup_blacklist(char *blacklist, char *entry)
{
    if(spasm_lookup_bl(blacklist, entry) == 1)
	printf("1\n");
    else
	printf("0\n");
}

void
get_blacklist(char *blacklist)
{
    struct spasm_list *first;
    struct spasm_list *cur;

    first = spasm_get_bl(blacklist);
    cur = first;
    while(cur != NULL) {
        if(cur->strval[0] != '\0' && (strcmp(cur->strval, "1") != 0)) {
	    if(strstr(cur->strval, "//")) printf("%s %s\n", cur->entry, cur->strval);         
	    else printf("%s // %s\n", cur->entry, cur->strval);
	} else printf("%s\n", cur->entry);
        first = cur->next;
        free(cur);
	cur = first;
    }
}

int
set_blacklist(char *blacklist)
{
    char *nl;
    char data[256];
    struct spasm_list *tmp, *first=NULL, *entries;

    while(fgets(data, sizeof(data), stdin) != NULL) {
	if((nl = strchr(data, '\n')) != NULL) *nl = '\0';
	if((nl = strchr(data, ' ')) != NULL || (nl = strchr(data, '\t')) != NULL) {
	    *nl = '\0';
	    nl++;
	}
	if(nl == NULL) nl = "1";
        if((tmp = calloc(sizeof(struct spasm_list), 1)) != NULL) {
            if(strlen(data) <= SPASM_DBT_MAX) {
                strncpy(tmp->entry, data, strlen(data));
		strncpy(tmp->strval, nl, strlen(nl));
                if (!first) {
                    first = entries = tmp;
                } else {                
                    entries->next = tmp;
                    entries = tmp;
                }
            } else {
                spasm_error(LOG_ERR, "Entry is too large");
                exit (1);
            }
        } else {
            spasm_error(LOG_ERR, "Error allocating memory for tmp: %m");
            exit (1);
        }
    }
    return (!spasm_set_bl(blacklist, first));
}

void
add_whitelist(char *whitelist, char *entry, char *val)
{
    spasm_insert_wl(whitelist, entry, val);
}

void
del_whitelist(char *whitelist, char *entry)
{
    spasm_delete_wl(whitelist, entry);
}

void
get_whitelist(char *whitelist)
{
    struct spasm_list *first;
    struct spasm_list *cur;

    first = spasm_get_wl(whitelist);
    cur = first;
    while(cur != NULL) {
        if(cur->strval[0] != '\0' && (strcmp(cur->strval, "1") != 0)) {
            if(strstr(cur->strval, "//")) printf("%s %s\n", cur->entry, cur->strval);
            else printf("%s // %s\n", cur->entry, cur->strval);
        } else printf("%s\n", cur->entry);
        first = cur->next;
        free(cur);
        cur = first;
    }
}

int
set_whitelist(char *whitelist)
{
    char *nl;
    char data[256];
    struct spasm_list *tmp, *first=NULL, *entries;

    while(fgets(data, sizeof(data), stdin) != NULL) {
        if((nl = strchr(data, '\n')) != NULL) *nl = '\0';
        if((nl = strchr(data, ' ')) != NULL || (nl = strchr(data, '\t')) != NULL) {
            *nl = '\0';
            nl++;
        }
        if(nl == NULL) nl = "1";
        if((tmp = calloc(sizeof(struct spasm_list), 1)) != NULL) {
            if(strlen(data) <= SPASM_DBT_MAX) {
                strncpy(tmp->entry, data, strlen(data));
                strncpy(tmp->strval, nl, strlen(nl));
                if (!first) {
                    first = entries = tmp;
                } else {
                    entries->next = tmp;
                    entries = tmp;
                }
            } else {
                spasm_error(LOG_ERR, "Entry is too large");
                exit (1);
            }
        } else {
            spasm_error(LOG_ERR, "Error allocating memory for tmp: %m");
            exit (1);
        }
    }
    return (!spasm_set_wl(whitelist, first));
}

void
lookup_filter(char *username, char *entry)
{
    int rc;
    
    rc = spasm_lookup_filter(username, entry);
    printf("%d\n", rc);
}

void
set_filter(char *username, char *entry, int val)
{
    spasm_set_filter(username, entry, val);
}

void
lookup_prefs(char *username, char *entry)
{
    char *rv = spasm_lookup_prefs(username, entry);
    if(rv != NULL) {
	printf("%s\n", rv);
	free(rv);
    }
}

void
set_prefs(char *username, char *entry, char *value)
{
    spasm_set_prefs(username, entry, value);
}

void
del_prefs(char *username, char *entry)
{
    spasm_del_prefs(username, entry);
}

int
delete_qfile(char *username, char *qfile)
{
    char qfilepath[FILENAME_MAX+1];
    struct flock lockinfo;
    int fd;
    
    snprintf(qfilepath, sizeof(qfilepath), "%s/users/%s/quarantine/%s", spasm_config->spasmroot, username, qfile);
    /* lock file to prevent corruption */
    if((fd = open(qfilepath, O_RDWR)) == -1) {
        fprintf(stderr, "Error opening qfile (%s): %m\n", qfilepath);
        return 1;
    }
    memset(&lockinfo, 0, sizeof(lockinfo));
    lockinfo.l_type = F_WRLCK;
    if(fcntl(fd, F_SETLK, &lockinfo) == -1) {
        fprintf(stderr, "Error locking qfile (%s): %m\n", qfilepath);
        close(fd);
        return 1;
    }
    unlink(qfilepath);
    close(fd);
    return 0;
}

void
view_log(char *username, char *log)
{
    struct spasm_list *first;
    struct spasm_list *cur;

    if(strcmp(log, "ALL") == 0) first = spasm_get_log(username, NULL, 1);
    else first = spasm_get_log(username, log, 0);

    while(first != NULL) {
	cur = first;
	printf("%s:%s\n", cur->entry, cur->strval);
	first = cur->next;
	free(cur);
    }
}

void
delete_log(char *username, char *log)
{
    if(strcmp(log, "ALL") == 0) spasm_delete_log(username, NULL, 1);
    else spasm_delete_log(username, log, 0);
}

void
show_setting(char *setting)
{
    if(strcasecmp(setting, "debug") == 0) printf("%d\n", spasm_config->debug);
    else if(strcasecmp(setting, "spasmuser") == 0 && spasm_config->spasmuser != NULL) printf("%s\n", spasm_config->spasmuser);
    else if(strcasecmp(setting, "defaultdomain") == 0 && spasm_config->defaultdomain != NULL) printf("%s\n", spasm_config->defaultdomain);
    else if(strcasecmp(setting, "domainlogging") == 0) printf("%d\n", spasm_config->domainlogging);
    else if(strcasecmp(setting, "domlogexceptions") == 0 && spasm_config->domlogexceptions != NULL) printf("%s\n", spasm_config->domlogexceptions);
    else if(strcasecmp(setting, "spasmroot") == 0 && spasm_config->spasmroot != NULL) printf("%s\n", spasm_config->spasmroot);
    else if(strcasecmp(setting, "stats") == 0) printf("%d\n", spasm_config->stats);
    else if(strcasecmp(setting, "tempdir") == 0 && spasm_config->tempdir != NULL) printf("%s\n", spasm_config->tempdir);
    else if(strcasecmp(setting, "pidfile") == 0 && spasm_config->pidfile != NULL) printf("%s\n", spasm_config->pidfile);
    else if(strcasecmp(setting, "socket") == 0 && spasm_config->socket != NULL) printf("%s\n", spasm_config->socket);
    else if(strcasecmp(setting, "clamav") == 0) printf("%d\n", spasm_config->clamav);
    else if(strcasecmp(setting, "sahost") == 0 && spasm_config->sahost != NULL) printf("%s\n", inet_ntoa(*spasm_config->sahost));
    else if(strcasecmp(setting, "saport") == 0) printf("%d\n", ntohs(spasm_config->saport));
    else if(strcasecmp(setting, "verifypasswd") == 0) printf("%d\n", spasm_config->verifypasswd);
    else if(strcasecmp(setting, "verifyaliases") == 0) printf("%d\n", spasm_config->verifyaliases);
    else if(strcasecmp(setting, "aliasesfile") == 0 && spasm_config->aliasesfile != NULL) printf("%s\n", spasm_config->aliasesfile);
    else if(strcasecmp(setting, "verifyspasm") == 0) printf("%d\n", spasm_config->verifyspasm);
    else if(strcasecmp(setting, "verifycustom") == 0) printf("%d\n", spasm_config->verifycustom);
    else if(strcasecmp(setting, "unknownusersthreshold") == 0) printf("%d\n", spasm_config->unknownusersthreshold);
    else if(strcasecmp(setting, "sendmail") == 0 && spasm_config->sendmail != NULL) printf("%s\n", spasm_config->sendmail);
    else if(strcasecmp(setting, "syslogrejections") == 0) printf("%d\n", spasm_config->syslogrejections);
    else if(strcasecmp(setting, "greylist") == 0 && spasm_config->greylist != NULL) printf("%s\n", spasm_config->greylist);
    else if(strcasecmp(setting, "greylistdelay") == 0) printf("%d\n", spasm_config->greylistdelay);
    else if(strcasecmp(setting, "greylistexpire") == 0) printf("%d\n", spasm_config->greylistexpire);
    else if(strcasecmp(setting, "greylistupdate") == 0) printf("%d\n", spasm_config->greylistupdate);
    else if(strcasecmp(setting, "greylistforcebl") == 0) printf("%d\n", spasm_config->greylistforcebl);
    else if(strcasecmp(setting, "mtime") == 0) printf("%li\n", spasm_config->mtime);
}

#define CMDLINE_HELP 128   /* largest ASCII value + 1 */
#define CMDLINE_VRSN 129

/* some getopt cmdline option stuff */
static struct option long_opts[] =
{   
    {"help",         0, 0, CMDLINE_HELP},
    {"version",      0, 0, CMDLINE_VRSN},
    {NULL, 0, 0, 0}
};

static void
bad_cmdline_syntax(void)
{
    fprintf(stderr, "Try `%s --help' for more information\n", progname);
    exit(1);
}

static void
version(void)
{   
    fprintf(stderr, "%s %s\n", progname, SPASM_VERSION);
    exit(0);
}

static void
cmdline_help(void)
{   
    fprintf(stderr,
"usage: %s command\n"
"       %s [--help --version]\n\n",
           progname, progname);

    fprintf(stderr,
"where:\n\n"
"--help                    display this message and exit\n"
"--version                 display program version number and exit\n"
"\n"
"Commands are as follows:\n\n"
"add_blacklist (blacklist) (entry) (value)\n"
"del_blacklist (blacklist) (entry)\n"
"get_blacklist (blacklist)\n"
"lookup_blacklist (blacklist) (entry)\n"
"set_blacklist (blacklist)\n"
"add_whitelist (whitelist) (entry) (value)\n"
"del_whitelist (whitelist) (entry)\n"
"get_whitelist (whitelist)\n"
"set_whitelist (whitelist)\n"
"lookup_filter (username) (entry)\n"
"set_filter (username) (entry) (value)\n"
"lookup_prefs (username) (entry)\n"
"set_prefs (username) (entry) (value)\n"
"del_prefs (username) (entry)\n"
"delete_qfile (username) (qfile)\n"
"view_log (username) (log)\n"
"delete_log (username) (log)\n"
"show_setting (setting)\n"
"\n"
"set_blacklist and set_whitelist read from STDIN\n"
"show_setting is a hook into %s\n\n", SPASM_CONF);

    exit(0);
}

int
main(int argc, char **argv)
{
    static char *command;
    char *args = "";
    char *slash = strrchr(argv[0], '/');

    progname = (slash) ? slash + 1 : argv[0];

    /* Command line processing */
    for(;;)
    {
       int rc;
        
        rc = getopt_long(argc, argv, args, long_opts, NULL);
        if (rc == -1)
           break;

       switch (rc)
       {   
           case CMDLINE_HELP:
               cmdline_help();
               /* NOT REACHED */
               break;
           case CMDLINE_VRSN:
               version();
               /* NOT REACHED */                                                                                                           
               break;                                                                                                                      
           case '?':
               bad_cmdline_syntax();
               /* NOT REACHED */
               break;
       }
    }

    if(argc < optind+1) bad_cmdline_syntax();

    command = argv[optind];

    if((spasm_config = spasm_cfg()) == NULL) {
        exit(1);                                                                                                              
    }

    if(strcasecmp(command, "add_blacklist") == 0) {
	if(argc < optind+3) bad_cmdline_syntax();
	else if(argc < optind+4) add_blacklist(argv[optind+1],argv[optind+2],"1");
	else add_blacklist(argv[optind+1], argv[optind+2], argv[optind+3]);
    } else if(strcasecmp(command, "del_blacklist") == 0) {
	if(argc < optind+3) bad_cmdline_syntax();
	del_blacklist(argv[optind+1], argv[optind+2]);
    } else if(strcasecmp(command, "get_blacklist") == 0) {
	if(argc < optind+2) bad_cmdline_syntax();
	get_blacklist(argv[optind+1]);
    } else if(strcasecmp(command, "lookup_blacklist") == 0) {
	if(argc < optind+3) bad_cmdline_syntax();
	lookup_blacklist(argv[optind+1], argv[optind+2]);
    } else if(strcasecmp(command, "set_blacklist") == 0) {
	if(argc < optind+2) bad_cmdline_syntax();
	set_blacklist(argv[optind+1]);
    } else if(strcasecmp(command, "add_whitelist") == 0) {
        if(argc < optind+3) bad_cmdline_syntax();
        else if(argc < optind+4) add_whitelist(argv[optind+1],argv[optind+2],"1");
        else add_whitelist(argv[optind+1], argv[optind+2], argv[optind+3]);
    } else if(strcasecmp(command, "del_whitelist") == 0) {
	if(argc < optind+3) bad_cmdline_syntax();
	del_whitelist(argv[optind+1], argv[optind+2]);
    } else if(strcasecmp(command, "get_whitelist") == 0) {
	if(argc < optind+2) bad_cmdline_syntax();
	get_whitelist(argv[optind+1]);
    } else if(strcasecmp(command, "set_whitelist") == 0) {
	if(argc < optind+2) bad_cmdline_syntax();
	set_whitelist(argv[optind+1]);
    } else if(strcasecmp(command, "lookup_filter") == 0) {
	if(argc < optind+3) bad_cmdline_syntax();
	lookup_filter(argv[optind+1], argv[optind+2]);
    } else if(strcasecmp(command, "set_filter") == 0) {
	if(argc < optind+4) bad_cmdline_syntax();
	set_filter(argv[optind+1], argv[optind+2], atoi(argv[optind+3]));
    } else if(strcasecmp(command, "lookup_prefs") == 0) {
	if(argc < optind+3) bad_cmdline_syntax();
	lookup_prefs(argv[optind+1], argv[optind+2]);
    } else if(strcasecmp(command, "set_prefs") == 0) {
	if(argc < optind+4) bad_cmdline_syntax();
	set_prefs(argv[optind+1], argv[optind+2], argv[optind+3]);
    } else if(strcasecmp(command, "del_prefs") == 0) {
	if(argc < optind+3) bad_cmdline_syntax();
	del_prefs(argv[optind+1], argv[optind+2]);
    } else if(strcasecmp(command, "delete_qfile") == 0) {
	if(argc < optind+3) bad_cmdline_syntax();
	delete_qfile(argv[optind+1], argv[optind+2]);
    } else if(strcasecmp(command, "view_log") == 0) {
	if(argc < optind+2) bad_cmdline_syntax();
	view_log(argv[optind+1], argv[optind+2]);
    } else if(strcasecmp(command, "delete_log") == 0) {
	if(argc < optind+3) bad_cmdline_syntax();
	delete_log(argv[optind+1], argv[optind+2]);
    } else if(strcasecmp(command, "show_setting") == 0) {
	if(argc < optind+2) bad_cmdline_syntax();
	show_setting(argv[optind+1]);
    } else bad_cmdline_syntax();

    exit(0);
}
