/*
**    spasm - The spasm anti-spam milter
**    Copyright (C) 2002-2005 Sean Hanson <aradia@theasylum.org>
**
**    This program is free software; you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation; either version 2 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program; if not, write to the Free Software
**    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef _SPASM_H
#define _SPASM_H

#include <netinet/in.h>

#ifdef HAVE_CONFIG_H
  #include "config.h"
#endif

/* Syslog Facility */
#define LOG_FACILITY		LOG_LOCAL6

/* Name of SA_REQUIRED_HITS filter */
#define SA_REQUIRED_HITS "SA_REQUIRED_HITS"

/* Filter status values */
#define SPASM_OFF		0
#define SPASM_TAG		1	/* TAG | LOG = 9 */
#define SPASM_QUARANTINE 	2	/* QUARANTINE | LOG = 10 */
#define SPASM_REJECT		4	/* REJECT | LOG = 12 */
#define SPASM_LOG		8

/* List types */
#define SPASM_BLACKLIST		1
#define SPASM_WHITELIST		2
#define SPASM_SYSTEM		3

#define SPASM_VERSION		"2.2"

/* Size limits per RFC 2821, Section 4.5.3.1 */
#define SMTP_LOCALPART_MAX 64
#define SMTP_DOMAIN_MAX 255
#define SMTP_PATH_MAX 320
#define SMTP_LINE_MAX 32767
/* End of RFC 2821 size limits */

#define NS_MAXMSG		65535

#define NIPQUAD(addr) \
        ((unsigned char *)&addr)[0], \
        ((unsigned char *)&addr)[1], \
        ((unsigned char *)&addr)[2], \
        ((unsigned char *)&addr)[3]

#define RNIPQUAD(addr) \
        ((unsigned char *)&addr)[3], \
        ((unsigned char *)&addr)[2], \
        ((unsigned char *)&addr)[1], \
        ((unsigned char *)&addr)[0]

struct spasmConfigStruct {
	int	debug;
	char*	spasmuser;
	char*	defaultdomain;
	int	domainlogging;
	char*	domlogexceptions;

	char*	spasmroot;
	char*	tempdir;
	char*	pidfile;
	char*	socket;

	int	clamav;

	struct in_addr* sahost;
	in_port_t saport;

	int	stats;

	int	verifypasswd;
	int	verifyaliases;
	char*	aliasesfile;
	int	verifyspasm;
	int	verifycustom;

	int	unknownusersthreshold;
	char*	sendmail;
	int	syslogrejections;

	char*	greylist;
	int	greylistdelay;
	int	greylistexpire;
	int	greylistupdate;
	int	greylistexpireinterval;
	int	greylistforcebl;

	char*	verifymx;
	int	verifymxexpire;
	int	verifymxexpireinterval;
	int	verifymxmaxattempts;

	int	defaultqexpire;
	int	defaultlexpire;

	int	spf;

	time_t	mtime;
};

void free_spasm_cfg(struct spasmConfigStruct *config);
struct spasmConfigStruct* spasm_cfg(void);
int verify_custom(char *username, char *domain);

#include "dbfunc.h"

#endif /* _SPASM_H */
