#include <arpa/inet.h>
#include <netinet/in.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include "spasm.h"

void spasm_error(int syslevel, char *errstr, ...);

/* Cleanup memory used by config struct */
void
free_spasm_cfg(struct spasmConfigStruct *config)
{
    if(config == NULL) return;

    if(config->spasmuser != NULL) free(config->spasmuser);
    if(config->defaultdomain != NULL) free(config->defaultdomain);
    if(config->domlogexceptions != NULL) free(config->domlogexceptions);
    if(config->spasmroot != NULL) free(config->spasmroot);
    if(config->tempdir != NULL) free(config->tempdir);
    if(config->pidfile != NULL) free(config->pidfile);
    if(config->socket != NULL) free(config->socket);
    if(config->sahost != NULL) free(config->sahost);
    if(config->aliasesfile != NULL) free(config->aliasesfile);
    if(config->sendmail != NULL) free(config->sendmail);
    if(config->greylist != NULL) free(config->greylist);
    if(config->verifymx != NULL) free(config->verifymx);

    free(config);
}

/* Read and return spasm.conf data */
struct spasmConfigStruct* spasm_cfg()
{
    FILE *fp;
    char buf[128];
    char key[128];
    char val[128];
    int seenthresh = 0;
    struct stat filestat;
    struct spasmConfigStruct *spasm_conf;

    /* stat file to get mtime */
    if(stat(SPASM_CONF, &filestat) < 0) {
	spasm_error(LOG_ERR, "(spasm_cfg) Unable to stat config file: %m");
	return NULL;
    }

    /* Open file for reading */
    if((fp = fopen(SPASM_CONF, "r")) == NULL) {
	spasm_error(LOG_ERR, "(spasm_cfg) Error opening config file: %m");
	return NULL;
    }
    /* Allocate memory... */
    if((spasm_conf = calloc(sizeof(struct spasmConfigStruct), 1)) == NULL) {
	spasm_error(LOG_ERR, "(spasm_cfg) Error allocating memory for spasm_conf: %m");
	return NULL;
    }
    /* And load the options! */
    while(fgets(buf, sizeof(buf), fp) != NULL) {
	/* Skip comments */
	if(buf[0] == '#') continue;

	/* Split key/value pair */
	if(sscanf(buf, "%127s %127s", key, val) != 2) continue;

	/* Check for key validity, add value to struct if valid */
	if(strcasecmp(key, "debug") == 0) {
	    if(strcasecmp(val, "on") == 0) spasm_conf->debug = 1;
	} else if(strcasecmp(key, "spasmuser") == 0) {
            spasm_conf->spasmuser = strdup(val);
            if(spasm_conf->spasmuser == NULL) {
                spasm_error(LOG_ERR, "(spasm_cfg) Error allocating memory for SpasmUser: %m");
                fclose(fp);
                return NULL;
            }
	} else if(strcasecmp(key, "defaultdomain") == 0) {
	    spasm_conf->defaultdomain = strdup(val);
	    if(spasm_conf->defaultdomain == NULL) {
		spasm_error(LOG_ERR, "(spasm_cfg) Error allocating memory for DefaultDomain: %m");
		fclose(fp);
		return NULL;
	    }
	} else if(strcasecmp(key, "domainlogging") == 0) {
	    if(strcasecmp(val, "on") == 0) spasm_conf->domainlogging = 1;
	} else if(strcasecmp(key, "domainloggingexceptions") == 0) {
	    spasm_conf->domlogexceptions = strdup(val);
	    if(spasm_conf->domlogexceptions == NULL) {
		spasm_error(LOG_ERR, "(spasm_cfg) Error allocating memory for DomainLoggingExceptions: %m");
		fclose(fp);
		return NULL;
	    }
	} else if(strcasecmp(key, "spasmroot") == 0) {
            spasm_conf->spasmroot = strdup(val);
            if(spasm_conf->spasmroot == NULL) {
                spasm_error(LOG_ERR, "(spasm_cfg) Error allocating memory for SpasmRoot: %m");
                fclose(fp);
                return NULL;
            }
        } else if(strcasecmp(key, "tempdir") == 0) {
            spasm_conf->tempdir = strdup(val);
            if(spasm_conf->tempdir == NULL) {
                spasm_error(LOG_ERR, "(spasm_cfg) Error allocating memory for TempDir: %m");
                fclose(fp);
                return NULL;
            }
	} else if(strcasecmp(key, "pidfile") == 0) {
            spasm_conf->pidfile = strdup(val);
            if(spasm_conf->pidfile == NULL) {
                spasm_error(LOG_ERR, "(spasm_cfg) Error allocating memory for PidFile: %m");
                fclose(fp);
                return NULL;
            }
        } else if(strcasecmp(key, "socket") == 0) {
            spasm_conf->socket = strdup(val);
            if(spasm_conf->socket == NULL) {
                spasm_error(LOG_ERR, "(spasm_cfg) Error allocating memory for Socket: %m");
                fclose(fp);
                return NULL;
            }
        } else if(strcasecmp(key, "clamav") == 0) {
	    if(strcasecmp(val, "on") == 0) spasm_conf->clamav = 1;
	} else if(strcasecmp(key, "spamassassinhost") == 0) {
	    if((spasm_conf->sahost = calloc(sizeof(struct in_addr), 1)) == NULL) {
		spasm_error(LOG_ERR, "(spasm_cfg) Error allocating memory for SpamAssassinHost: %m");
		fclose(fp);
		return NULL;
	    }
	    if(inet_aton(val, spasm_conf->sahost) == 0) {
		spasm_error(LOG_ERR, "(spasm_cfg) Invalid SpamAssassin Host: %s", val);
		fclose(fp);
		return NULL;
	    }
	} else if(strcasecmp(key, "spamassassinport") == 0) {
	    spasm_conf->saport = htons(atoi(val));
	} else if(strcasecmp(key, "verifyuserspasswd") == 0) {
	    if(strcasecmp(val, "on") == 0) spasm_conf->verifypasswd = 1;
        } else if(strcasecmp(key, "verifyusersaliases") == 0) {
            if(strcasecmp(val, "on") == 0) spasm_conf->verifyaliases = 1;
        } else if(strcasecmp(key, "verifyusersspasm") == 0) {
            if(strcasecmp(val, "on") == 0) spasm_conf->verifyspasm = 1;
        } else if(strcasecmp(key, "verifyuserscustom") == 0) {
            if(strcasecmp(val, "on") == 0) spasm_conf->verifycustom = 1;
	} else if(strcasecmp(key, "aliasesfile") == 0) {
            spasm_conf->aliasesfile = strdup(val);
            if(spasm_conf->aliasesfile == NULL) {
                spasm_error(LOG_ERR, "(spasm_cfg) Error allocating memory for AliasesFile: %m");
                fclose(fp);
                return NULL;
            }
	} else if(strcasecmp(key, "sendmail") == 0) {
            spasm_conf->sendmail = strdup(val);
            if(spasm_conf->sendmail == NULL) {
                spasm_error(LOG_ERR, "(spasm_cfg) Error allocating memory for Sendmail: %m");
                fclose(fp);
                return NULL;
            }
	} else if(strcasecmp(key, "unknownusersthreshold") == 0) {
	    spasm_conf->unknownusersthreshold = atoi(val);
	    seenthresh = 1;
	} else if(strcasecmp(key, "syslogrejections") == 0) {
	    if(strcasecmp(val, "on") == 0) spasm_conf->syslogrejections = 1;
	} else if(strcasecmp(key, "greylist") == 0) {
	    spasm_conf->greylist = strdup(val);
	    if(spasm_conf->greylist == NULL) {
		spasm_error(LOG_ERR, "(spasm_cfg) Error allocating memory for Greylist: %m");
		fclose(fp);
		return NULL;
	    }
	} else if(strcasecmp(key, "greylistdelay") == 0) {
	    spasm_conf->greylistdelay = atoi(val) * 60; /* minutes -> seconds */
	} else if(strcasecmp(key, "greylistexpire") == 0) {
	    spasm_conf->greylistexpire = atoi(val) * 3600; /* hours -> seconds */
	} else if(strcasecmp(key, "greylistupdate") == 0) {
	    spasm_conf->greylistupdate = atoi(val) * 86400; /* days -> seconds */
	} else if(strcasecmp(key, "greylistexpireinterval") == 0) {
	    spasm_conf->greylistexpireinterval = atoi(val) * 60; /* minutes -> seconds */
	} else if(strcasecmp(key, "greylistforcebl") == 0) {
	    if(strcasecmp(val, "on") == 0) spasm_conf->greylistforcebl = 1;
	} else if(strcasecmp(key, "defaultquarantineexpire") == 0) {
	    spasm_conf->defaultqexpire = atoi(val);
	} else if(strcasecmp(key, "defaultlogexpire") == 0) {
	    spasm_conf->defaultlexpire = atoi(val);
	} else if(strcasecmp(key, "verifymx") == 0) {
	    spasm_conf->verifymx = strdup(val);
	    if(spasm_conf->verifymx == NULL) {
		spasm_error(LOG_ERR, "(spasm_cfg) Error allocating memory for VerifyMX: %m");
		fclose(fp);
		return NULL;
	    }
	} else if(strcasecmp(key, "verifymxexpire") == 0) {
	    spasm_conf->verifymxexpire = atoi(val) * 60; /* minutes -> seconds */
	} else if(strcasecmp(key, "verifymxexpireinterval") == 0) {
	    spasm_conf->verifymxexpireinterval = atoi(val) * 60; /* minutes -> seconds */
	} else if(strcasecmp(key, "verifymxmaxattempts") == 0) {
	    spasm_conf->verifymxmaxattempts = atoi(val);
	} else if(strcasecmp(key, "stats") == 0) {
	    if(strcasecmp(val, "on") == 0) spasm_conf->stats = 1;
	} else if(strcasecmp(key, "spf") == 0) {
	    if(strcasecmp(val, "on") == 0) spasm_conf->spf = 1;
	} else {
	    spasm_error(LOG_WARNING, "(spasm_cfg) Unknown configuration option: %s", key);
	}
    }
    fclose(fp);

    /* Now add mtime to spasm_conf */
    spasm_conf->mtime = filestat.st_mtime;

    /* Now verify that we have all the needed options */
    if(spasm_conf->spasmuser == NULL) {
        spasm_error(LOG_ERR, "(spasm_cfg) SpasmUser not found");
        return NULL;
    }
    if(spasm_conf->defaultdomain == NULL) {
	spasm_error(LOG_ERR, "(spasm_cfg) DefaultDomain not found");
	return NULL;
    }
    if(spasm_conf->spasmroot == NULL) {
        spasm_error(LOG_ERR, "(spasm_cfg) SpasmRoot not found");
        return NULL;
    }
    if(spasm_conf->tempdir == NULL) {
	spasm_error(LOG_ERR, "(spasm_cfg) TempDir not found");
	return NULL;
    }
    if(spasm_conf->pidfile == NULL) {
        spasm_error(LOG_ERR, "(spasm_cfg) PidFile not found");
        return NULL;
    }
    if(spasm_conf->socket == NULL) {
	spasm_error(LOG_ERR, "(spasm_cfg) Socket not found");
	return NULL;
    }
    if(spasm_conf->sahost != NULL && spasm_conf->saport == 0) {
	spasm_error(LOG_ERR, "(spasm_cfg) SpamAssassinPort not found");
	return NULL;
    }
    if(spasm_conf->aliasesfile == NULL && spasm_conf->verifyaliases) {
        spasm_error(LOG_ERR, "(spasm_cfg) AliasesFile not found");
        return NULL;
    }
    if(spasm_conf->sendmail == NULL) {
        spasm_error(LOG_ERR, "(spasm_cfg) Sendmail not found");
        return NULL;
    }
    if(spasm_conf->greylist != NULL) {
        if(spasm_conf->greylistdelay == 0) spasm_conf->greylistdelay = 180;
        if(spasm_conf->greylistexpire == 0) spasm_conf->greylistexpire = 28800;
        if(spasm_conf->greylistupdate == 0) spasm_conf->greylistupdate = 604800;
	if(spasm_conf->greylistexpireinterval == 0) spasm_conf->greylistexpireinterval = 900;
    }
    if(spasm_conf->verifymx != NULL) {
	if(spasm_conf->verifymxexpire == 0) spasm_conf->verifymxexpire = 3600;
	if(spasm_conf->verifymxexpireinterval == 0) spasm_conf->verifymxexpireinterval = 900;
    }
    if(spasm_conf->defaultqexpire == 0) spasm_conf->defaultqexpire = 14;
    if(spasm_conf->defaultlexpire == 0) spasm_conf->defaultlexpire = 14;
    if(!seenthresh) spasm_conf->unknownusersthreshold = 5;

    return spasm_conf;
}
