/*
**    spasm - The spasm anti-spam milter
**    Copyright (C) 2002-2005 Sean Hanson <aradia@theasylum.org>
**
**    This program is free software; you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation; either version 2 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program; if not, write to the Free Software
**    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <ctype.h>
#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "spasm.h"
#include "getopt-portable.h"

struct spasmConfigStruct *spasm_config;
static char *progname;

void spasm_error(int syslevel, char *errstr, ...)
{
    va_list arglist;
#ifdef SUNOS
    char *at, message[1024], temp[1024];

    strlcpy(message, errstr, sizeof(message));
    while ((at=strstr(message, "%m")) != NULL) {
        *at = '\0';
        strlcpy(temp, at+2, sizeof(temp));
        strlcat(message, strerror(errno), sizeof(message));
        strlcat(message, temp, sizeof(message));
    }
#endif /* SUNOS */

    va_start(arglist, errstr);
#ifdef SUNOS
    vfprintf(stderr, message, arglist);
#else
    vfprintf(stderr, errstr, arglist);
#endif /* SUNOS */
    fprintf(stderr, "\n");	/* spasm_error args aren't terminated by \n */
    va_end(arglist);
}

#define CMDLINE_HELP 128   /* largest ASCII value + 1 */
#define CMDLINE_VRSN 129

/* some getopt cmdline option stuff */
static struct option long_opts[] =
{
    {"help",         0, 0, CMDLINE_HELP},
    {"version",      0, 0, CMDLINE_VRSN},
    {NULL, 0, 0, 0}
};

static void
bad_cmdline_syntax(void)
{
    fprintf(stderr, "Try `%s --help' for more information\n", progname);
    exit(1);
}

static void
version(void)
{
    fprintf(stderr, "%s %s\n", progname, SPASM_VERSION);
    exit(0);
}

static void
cmdline_help(void)
{
    fprintf(stderr,
            "usage: %s [--help --version] <dbname>\n\n",
           progname);

    fprintf(stderr,
"where:\n\n"
"<dbname>                  is the name of a blacklist database to be created.\n"
"                          Note that database input is read from stdin.\n\n"
"--help                    display this message and exit\n"
"--version                 display program version number and exit\n");

    exit(0);
}

int main(int argc, char **argv)
{
    char data[256], *val, *dbname;
    struct spasm_list *entries = NULL, *tmp;
    struct spasm_list *first = NULL;
    const char *args = "c:";

    char *slash = strrchr(argv[0], '/');

    progname = (slash) ? slash + 1 : argv[0];

    /* command line processing first */
    for (;;)
    {
       int rc;

        rc = getopt_long(argc, argv, args, long_opts, NULL);
        if (rc == -1)
           break;

       switch (rc)
       {
           case CMDLINE_HELP:
               cmdline_help();
               /* NOT REACHED */
               break;
           case CMDLINE_VRSN:
               version();
               /* NOT REACHED */
               break;
           case '?':
               bad_cmdline_syntax();
               /* NOT REACHED */
               break;
       }
    }
    if (argc < optind + 1) {
        /* required argument missing */
        fprintf(stderr, "%s: Missing required arguments\n", progname);
        bad_cmdline_syntax();
        /* NOT REACHED */
    } else if(argc > optind + 1) {
        /* too many arguments */
	fprintf(stderr, "%s: Too many arguments\n", progname);
	bad_cmdline_syntax();
	/* NOT REACHED */
    }

    dbname = argv[optind];

    if((spasm_config = spasm_cfg()) == NULL) {
	exit(1);
    }

    while(fgets(data, sizeof(data), stdin) != NULL) {
        if (data[0] == '#') continue;
        val = data;
        while(isspace(*val) == 0) val++;
        *val = '\0';
        if((tmp = calloc(sizeof(struct spasm_list), 1)) != NULL) {
            if(strlen(data) <= SPASM_DBT_MAX) {
                strncpy(tmp->entry, data, strlen(data));
                if (!first) {
                    first = entries = tmp;                                                                  
                } else {                                                                                    
                    entries->next = tmp;
                    entries = tmp;
                }
            } else {
                spasm_error(LOG_ERR, "Entry is too large");
                exit (1);
            }
        } else {
            spasm_error(LOG_ERR, "Error allocating memory for tmp: %m");
            exit (1);
        }
    }
    return (!spasm_set_db(dbname, first));
}
