package spasm::tools;
use strict;
use MIME::Base64;
use MIME::Parser;
use Time::Local;
use vars qw(%months @month_en @wday_en %tzoffset);

#
# Tools for spasm webmail
#

# escapeURL from OpenWebmail
sub escapeURL {
    my $toencode = shift;
    return undef if (!defined($toencode));
    return unless $toencode;
    $toencode=~s/([^a-zA-Z0-9_.-])/uc sprintf("%%%02x",ord($1))/eg;
    return $toencode;
}

# email2nameaddr from OpenWebmail
sub email2nameaddr {    # name, addr are guarentee to not null
   my ($name, $address)=_email2nameaddr($_[0]);
   if ($name eq "") {
      $name=$address; $name=~s/\@.*$//;
      $name=$address if (length($name)<=2);
   }
   return($name, $address);
}

# _email2nameaddr from OpenWebmail
sub _email2nameaddr {   # name may be null
   my $email=$_[0];
   my ($name, $address);
   
   if ($email=~/^\s*"?<?(.+?)>?"?\s*<(.*)>$/) {
      $name = $1; $address = $2;                                                                                                                       
   } elsif ($email=~/<?(.*?@.*?)>?\s+\((.+?)\)/) {                                                                                                     
      $name = $2; $address = $1;
   } elsif ($email=~/<(.+)>/) {
      $name = ""; $address = $1;
   } elsif ($email=~/(.+)/) {
      $name = "" ; $address = $1;
   }
   $name=~s/^\s+//; $name=~s/\s+$//;
   $address=~s/^\s+//; $address=~s/\s+$//;
   return($name, $address);
}

# str2html from OpenWebmail
sub str2html {
   my $t=$_[0];

   $t=~s/&#(\d\d\d\d);/ESCAPE_UNICODE_$1/g;
   $t=~s/&/ESCAPE_AMP/g;

   $t=~s/\"/&quot;/g;
   $t=~s/</&lt;/g;
   $t=~s/>/&gt;/g;

   $t=~s/ESCAPE_AMP/&amp;/g;
   $t=~s/ESCAPE_UNICODE_(\d\d\d\d)/&#$1;/g;

   return($t);
}

# text2html from OpenWebmail
sub text2html {
   my $t=$_[0];

   $t=~s/&#(\d\d\d+);/ESCAPE_UNICODE_$1/g;
   $t=~s/&/ESCAPE_AMP/g;

   $t=~s/\"/ &quot; /g;
   $t=~s/</ &lt; /g;
   $t=~s/>/ &gt; /g;

   $t=~s!(https?|ftp|mms|nntp|news|gopher|telnet)://([\w\d\-\.]+?/?[^\s\(\)\<\>\x80-\xFF]*[\w/])([\b|\n| ]*)!<a href="$1://$2" target="_blank">$1://$2</a>$3!gs;
   $t=~s!([\b|\n| ]+)(www\.[\w\d\-\.]+\.[\w\d\-]{2,4})([\b|\n| ]*)!$1<a href="http://$2" target="_blank">$2</a>$3!igs;
   $t=~s!([\b|\n| ]+)(ftp\.[\w\d\-\.]+\.[\w\d\-]{2,4})([\b|\n| ]*)!$1<a href="ftp://$2" target="_blank">$2</a>$3!igs;

   # remove the blank inserted just now
   $t=~s/ (&quot;|&lt;|&gt;) /$1/g;
   
   $t=~s/ {2}/ &nbsp;/g;
   $t=~s/\t/ &nbsp;&nbsp;&nbsp;&nbsp;/g;
   $t=~s/\n/ <BR>\n/g;
   
   $t=~s/ESCAPE_AMP/&amp;/g;
   $t=~s/ESCAPE_UNICODE_(\d\d\d+)/&#$1;/g;
   
   return($t);
}

# _pre2html from OpenWebmail
sub _pre2html {
   my $t=$_[0];

#   $t=~s/\"/&quot;/g; $t=~s/</&lt;/g; $t=~s/>/&gt;/g;
   $t=~s!</(?:p|div|table|th|tr)> *\r?\n!</$1>ESCAPE_NEWLINE!ig;
   $t=~s!\n!<br>\n!g;
   $t=~s!ESCAPE_NEWLINE!\n!ig;
   return($t);
}

# html2text from OpenWebmail
sub html2text {
   my $t=$_[0];
   
   # turn <pre>...</pre> into pure html with <br>
   $t=~s|<!--+-->|\n|isg;
   $t=~s!\s*<pre[^\<\>]*?>(.*?)</pre>\s*!_pre2html($1)!iges;
   
   $t=~s!@!  !g;       # clean chinese big5 space char
   $t=~s!&nbsp;! !g;
   $t=~s![ \t]+! !g;
   $t=~s!^\s+!!mg; $t=~s!\s+$! !mg;
   $t=~s![\r\n]+!!g;
   
   $t=~s|<!--.*?-->||sg;
   $t=~s!<style(?: [^\<\>]*)?>.*?</style>!!isg;
   $t=~s!<script(?: [^\<\>]*)?>.*?</script>!!isg;
   $t=~s!<noframes(?: [^\<\>]*)?>.*?</noframes>!!isg;
   $t=~s!<i?frame[^\<\>]* src="?([^\<\>\s\"]*)"?[^\<\>]*>(.*?</iframe>)?!\n$1\n!isg;
   
   $t=~s!<p(?: [^\<\>]*)?>!ESCAPE_P!ig;
   $t=~s!<div(?: [^\<\>]*)?>!ESCAPE_DIV!ig;
   
   # this should be processed before </td> <-> space replacement,
   # or </p>|</div>|\s before </td> won't be found
   $t=~s!<(?:span|font|b|i|a)(?: [^\<\>]*)?>!!isg;
   $t=~s!</(?:th|tr|span|font|b|i|a)>!!ig;
   
   # this should be processed before <table> <-> \n replacement,
   # or it will eat \n of <table> in recursive table
   $t=~s!<td(?: [^\<\>]*)?>(?:ESCAPE_P|ESCAPE_DIV|</p>|</div>|\s)*</td>!!isg;
   $t=~s!<td(?: [^\<\>]*)?>(?:ESCAPE_P|ESCAPE_DIV|\s)*! !isg;
   $t=~s!(?:</p>|</div>|\s)*</td>! !isg;

   $t=~s!<(?:table|tbody|th|tr)(?: [^\<\>]*)?>!\n!isg;
   $t=~s!</(?:table|tbody)>!\n!ig;

   $t=~s!<(?:ol|ul)(?: [^\<\>]*)?>!\n!ig;
   $t=~s!</(?:ol|ul)>!\n!ig;
   $t=~s!(?:</p>|</div>|\s)*<li>(?:ESCAPE_P|ESCAPE_DIV|\s)*!\n* !isg;

   $t=~s!ESCAPE_P\s*(?:</p>)*!\n\n!isg;
   $t=~s!</p>!\n\n!ig;

   $t=~s!ESCAPE_DIV\s*(?:</div>)*!\n\n!isg;
   $t=~s!</div>!\n\n!ig;
   
   $t=~s!<select(?: [^\<\>]*)?>(<option(?: [^\<\>]*)?>)?!(!isg;
   $t=~s!</select>!)!ig;
   $t=~s!<option(?: [^\<\>]*)?>!,!isg;
   $t=~s!<input[^\<\>]* type=['"]?radio['"]?[^\<\>]*>! *!isg;
   
   $t=~s!</?title>!\n\n!ig;
   $t=~s!<br ?/?>!\n!ig;
   $t=~s!<hr(?: [^\<\>]*)?>!\n-----------------------------------------------------------------------\n!ig;
   
   $t=~s!<[^\<\>]+?>!!sg;
   
   $t=~s!&lt;!<!g;
   $t=~s!&gt;!>!g;
   $t=~s!&amp;!&!g;
   $t=~s!&quot;!\"!g;
   $t=~s!&copy;!(C)!g;
   
   $t=~s!^\s+!!;
   $t=~s!(?:[ |\t]*\n){2,}!\n\n!sg;
   
   return($t);
}

sub formatHeader {
    my $fullheaders = shift;
    my @headers = @_;
    my $formatted_hdrs = "";
    my $button_val;
    my $button_text;
    my $button_title;

    unless(defined($fullheaders) and @headers) { return; }

    if($fullheaders) {
	$button_val = 0;
	$button_text = "Simple Headers";
	$button_title = "View Simple Headers";
    } else {
	$button_val = 1;
	$button_text = "Full Headers";
	$button_title = "View Full Headers";
    }

    my $firsthdr = 1;
    for my $i (0..scalar(@headers)-1) {
	my $hdr = $headers[$i];
	$hdr =~ s/[\n|\r]//g;
	unless($hdr) { next; }
	$hdr = decode_mimewords($hdr);
	$hdr = str2html($hdr);
	my $f; my $v;
	if($hdr =~ /^(\S+):\s+(.+)$/) {
	    $f = $1; $v = $2;
	} elsif($hdr =~ /^(\S+):\s+$/) {
	    $f = $1; $v = "";
	} else {
	    $f = undef; $v = undef;
	}
	if($fullheaders < 1) {
	    if(!defined $f or !defined $v) { next; }
	    unless($f =~ /^(?:date|from|reply-to|to|cc|bcc|subject)$/i) { next; }
	}
	if(defined $f and defined $v) {
	    if($fullheaders < 0) {
		$formatted_hdrs .= qq|<tr><td bgcolor="#DDDDDD"><b>$f:</b> $v</td></tr>\n|;
	    } elsif($firsthdr) {
		$formatted_hdrs .= qq|<tr><td bgcolor="#DDDDDD"><b>$f:</b> $v</td><td bgcolor="#DDDDDD" align=right><a href='javascript: toggleHeaders("$button_val");' title='$button_title'>$button_text</a></td></tr>\n|;
	    } else {
		$formatted_hdrs .= qq|<tr><td colspan=2 bgcolor="#DDDDDD"><b>$f:</b> $v</td></tr>\n|;
	    }
	} else {
	    $formatted_hdrs .= qq|<tr><td colspan=2 bgcolor="#DDDDDD">&nbsp;&nbsp;&nbsp;&nbsp;$hdr</td></tr>\n|;
	}
	if($firsthdr) { $firsthdr = 0; }
	if($fullheaders < 1) {
            while(defined($headers[$i+1]) and $headers[$i+1] =~ /^\s+/) {
                $i++;
                $hdr = $headers[$i];
                $hdr =~ s/[\n|\r]//g;
                unless($hdr) { next; }
                $hdr = decode_mimewords($hdr);
                $hdr = str2html($hdr);
                if($fullheaders < 0) {
                    $formatted_hdrs .= qq|<tr><td bgcolor="#DDDDDD">&nbsp;&nbsp;&nbsp;&nbsp;$hdr</td></tr>\n|;
                } else {
                    $formatted_hdrs .= qq|<tr><td colspan=2 bgcolor="#DDDDDD">&nbsp;&nbsp;&nbsp;&nbsp;$hdr</td></tr>\n|;                               
                }                                                                                                                                      
            }
	}
    }
    return $formatted_hdrs;
}

# _decode_Q STRING
#     Private: used by _decode_header() to decode "Q" encoding, which is
#     almost, but not exactly, quoted-printable.  :-P
sub _decode_Q {
    my $str = shift;
    $str =~ s/_/\x20/g;                                # RFC-1522, Q rule 2
    $str =~ s/=([\da-fA-F]{2})/pack("C", hex($1))/ge;  # RFC-1522, Q rule 1
    $str;
}

# _decode_B STRING
#     Private: used by _decode_header() to decode "B" encoding.
sub _decode_B {
    my $str = shift;
    decode_base64($str);
}

sub decode_mimewords {
    my $encstr = shift;
    return unless $encstr;
    my %params = @_;
    my @tokens;
    $@ = '';           ### error-return

    ### Collapse boundaries between adjacent encoded words:
    $encstr =~ s{(\?\=)\s*(\=\?)}{$1$2}gs;
    pos($encstr) = 0;
    ### print STDOUT "ENC = [", $encstr, "]\n";

    ### Decode:
    my ($charset, $encoding, $enc, $dec);
    while (1) {
        last if (pos($encstr) >= length($encstr));
        my $pos = pos($encstr);               ### save it

        ### Case 1: are we looking at "=?..?..?="?
        if ($encstr =~    m{\G             # from where we left off..
                            =\?([^?]*)     # "=?" + charset +
                             \?([bq])      #  "?" + encoding +
                             \?([^?]+)     #  "?" + data maybe with spcs +
                             \?=           #  "?="
                            }xgi) {
            ($charset, $encoding, $enc) = ($1, lc($2), $3);
            $dec = (($encoding eq 'q') ? _decode_Q($enc) : _decode_B($enc));
            push @tokens, [$dec, $charset];
            next;
        }
        
        ### Case 2: are we looking at a bad "=?..." prefix?
        ### We need this to detect problems for case 3, which stops at "=?":
        pos($encstr) = $pos;               # reset the pointer.
        if ($encstr =~ m{\G=\?}xg) {
            $@ .= qq|unterminated "=?..?..?=" in "$encstr" (pos $pos)\n|;
            push @tokens, ['=?'];
            next;
        }

        ### Case 3: are we looking at ordinary text?
        pos($encstr) = $pos;               # reset the pointer.
        if ($encstr =~ m{\G                # from where we left off...
                         ([\x00-\xFF]*?    #   shortest possible string,
                          \n*)             #   followed by 0 or more NLs,
                         (?=(\Z|=\?))      # terminated by "=?" or EOS
                        }xg) {
            length($1) or die "mime: empty token";
            push @tokens, [$1];
            next;
        }

        ### Case 4: bug!
        die "mime: unexpected case:\n($encstr) pos $pos";
    }                                                                                               
    return (wantarray ? @tokens : join('',map {$_->[0]} @tokens));
}

#
# From OpenWebmail: modules/htmlrender.pl
#
sub html4nobase {
   my $html=$_[0];
   my $urlbase;

   if ( $html =~ m!\<base\s+href\s*=\s*"?([^\<\>]*?)"?\>!i ) {
      $urlbase=$1;
      $urlbase=~s!/[^/]+$!/!;
   }
   unless(defined $urlbase) { $urlbase = ""; }

   $html =~ s!\<base\s+([^\<\>]*?)\>!!gi;
   if ( ($urlbase ne "") && ($urlbase !~ /^file:/) ) {
      $html =~ s!(\<a\s+href|background|src|method|action)(=\s*"?)!$1$2$urlbase!gi;
      # recover links that should not be changed by base directive
      $html =~ s!\Q$urlbase\E(http://|https://|ftp://|mms://|cid:|mailto:|#)!$1!gi;
   }
   return($html);
}

my @jsevents=('onAbort', 'onBlur', 'onChange', 'onClick', 'onDblClick',
              'onDragDrop', 'onError', 'onFocus', 'onKeyDown', 'onKeyPress',
              'onKeyUp', 'onLoad', 'onMouseDown', 'onMouseMove', 'onMouseOut',
              'onMouseOver', 'onMouseUp', 'onMove', 'onReset', 'onResize',
              'onSelect', 'onSubmit', 'onUnload', 'window.open');

sub html4link {
   my $html=$_[0];
   $html=~s/(<a\s+[^\<\>]*?>)/_link_target_blank($1)/igems;
   return($html);
}

sub _link_target_blank {
   my $link=$_[0];
   if ($link =~ /(?:target=|javascript:|href="?#)/i ) {
      return($link);
   }
   $link=~s/<a\s+([^\<\>]*?)>/<a $1 target=_blank>/is;
   return($link);
}

sub html4noframe {
   my $html=$_[0];
   $html=~s/(<frame\s+[^\<\>]*?>)/_frame2iframe($1)/igems;
   return($html);
}

sub _frame2iframe {
   my $frame=$_[0];
   return "" if ( $frame!~/src=/i );
   $frame=~s/<frame /<iframe width="100%" height="250" /is;
   $frame.=qq|</iframe>|;
   return($frame);
}

sub html4disablejs {
   my $html=$_[0];
   foreach my $event (@jsevents) {
      $html=~s/$event/_$event/isg;
   }
   $html=~s/<script([^\<\>]*?)>/<disable_script$1>\n<!--\n/isg;
   $html=~s/<!--\s*<!--/<!--/isg;
   $html=~s/<\/script>/\n\/\/-->\n<\/disable_script>/isg;
   $html=~s/\/\/-->\s*\/\/-->/\/\/-->/isg;
   $html=~s/<([^\<\>]*?)javascript:([^\<\>]*?)>/<$1disable_javascript:$2>/isg;

   return($html);
}

sub html4disableembcode {
   my $html=$_[0];
   foreach my $tag (qw(embed applet object)) {
      $html=~s!<\s*$tag([^\<\>]*?)>!<disable_$tag$1>!isg;
      $html=~s!<\s*/$tag([^\<\>]*?)>!</disable_$tag$1>!isg;
   }
   $html=~s!<\s*param ([^\<\>]*?)>!<disable_param $1>!isg;
   return($html);
}

sub html4disableemblink {
   my ($html, $range, $blankimg)=@_;
   $html=~s!(src|background)\s*=\s*("?https?://[\w\.\-]+?/?[^\s<>]*)([\b|\n| ]*)!_clean_emblink($1,$2,$3,$range,$blankimg)!egis;
   return($html);
}

sub _clean_emblink {
   my ($type, $url, $end, $range, $blankimg)=@_;
   if ($url !~ /\Q$ENV{'HTTP_HOST'}\E/is) { # non-local URL found
      if ($range eq 'cgionly' && $url=~/\?/s) {
         $url=~s/["']//g;
         return(qq|border="1" |.
                qq|$type="$blankimg" |. # blank img url
                qq|alt="Embedded CGI removed by webmail.\n$url" |.
                qq|onclick="window.open('$url', '_extobj');" |.
                $end);
      } elsif ($range eq 'all') {
         $url=~s/["']//g;
         return(qq|border="1" |.
                qq|$type="$blankimg" |. # blank img url
                qq|alt="Embedded link removed by webmail.\n$url" |.
                qq|onclick="window.open('$url', '_extobj');" |.
                $end);
      }
   }
   return("$type=$url".$end);
}

sub html2table {
   my $html=_htmlclean($_[0]);
   $html =~ s#\<body([^\<\>]*?)\>#<table width=100% border=0 cellpadding=2 cellspacing=0 $1><tr><td>#is;
   $html =~ s#\</body\>#</td></tr></table>#i;
   return $html;
}

sub _htmlclean {
   my $html=$_[0];

   $html =~ s#<!doctype[^\<\>]*?>##i;
   $html =~ s#<html[^\<\>]*?>##i;
   $html =~ s#</html>##i;
   $html =~ s#<head>.*?</head>##is;
   $html =~ s#<head>##i;
   $html =~ s#</head>##i;
   $html =~ s#<meta[^\<\>]*?>##gi;
   $html =~ s#<!--.*?-->##gis;
   $html =~ s#<style[^\<\>]*?>#\n<!-- style begin\n#gi;
   $html =~ s#</style>#\nstyle end -->\n#gi;
   $html =~ s#<[^\<\>]*?stylesheet[^\<\>]*?>##gi;
   $html =~ s#(<div[^\<\>]*?)position\s*:\s*absolute\s*;([^\<\>]*?>)#$1$2#gi;
   return($html);
}

#
# From OpenWebmail: modules/datetime.pl
#
%months = qw(Jan 1 Feb 2 Mar 3 Apr 4  May 5  Jun 6
             Jul 7 Aug 8 Sep 9 Oct 10 Nov 11 Dec 12);
@month_en = qw(Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec);
@wday_en =  qw(Sun Mon Tue Wed Thu Fri Sat);

%tzoffset = qw(
    ACDT +1030  ACST +0930  ADT  -0300  AEDT +1100  AEST +1000  AHDT -0900
    AHST -1000  AST  -0400  AT   -0200  AWDT +0900  AWST +0800  AZST +0400
    BAT  +0300  BDST +0200  BET  -1100  BST  -0300  BT   +0300  BZT2 -0300
    CADT +1030  CAST +0930  CAT  -1000  CCT  +0800  CDT  -0500  CED  +0200
    CEST +0200  CET  +0100  CST  -0600
    EAST +1000  EDT  -0400  EED  +0300  EET  +0200  EEST +0300  EST  -0500
    FST  +0200  FWT  +0100
    GMT  +0000  GST  +1000
    HDT  -0900  HST  -1000
    IDLE +1200  IDLW -1200  IST  +0530  IT   +0330
    JST  +0900  JT   +0700
    MDT  -0600  MED  +0200  MET  +0100  MEST +0200  MEWT +0100  MST  -0700
    MT   +0800
    NDT  -0230  NFT  -0330  NT   -1100  NST  +0630  NZ   +1100  NZST +1200
    NZDT +1300  NZT  +1200
    PDT  -0700  PST  -0800
    ROK  +0900                                                                                                                                        
    SAD  +1000  SAST +0900  SAT  +0900  SDT  +1000  SST  +0200  SWT  +0100                                                                            
    USZ3 +0400  USZ4 +0500  USZ5 +0600  USZ6 +0700  UT   +0000  UTC  +0000
    UZ10 +1100
    WAT  -0100  WET  +0000  WST  +0800
    YDT  -0800  YST  -0900
    ZP4  +0400  ZP5  +0500  ZP6  +0600);

sub timeoffset2seconds {
   my $seconds=0;
   if ($_[0]=~/^[+\-]?(\d\d)(\d\d)$/) { # $_[0] is timeoffset
      $seconds=($1*60+$2)*60;                                                                                                                        
      $seconds*=-1 if ($_[0]=~/^\-/);
   }
   return($seconds);
}

sub seconds2timeoffset {
   my $seconds=abs($_[0]);
   return(sprintf( "%s%02d%02d",
        ($_[0]>=0)?'+':'-', int($seconds/3600), int(($seconds%3600)/60) ));
}

sub seconds2array {
   return gmtime($_[0]);
}

sub array2seconds {
   my ($sec,$min,$hour, $d,$m,$y)=@_;
   # Make sure everything's numeric
   unless($sec) { $sec = 0; }
   unless($min) { $min = 0; }
   unless($hour) { $hour = 0; }
   unless($d) { $d = 0; }
   unless($m) { $m = 0; }
   unless($y) { $y = 0; }

   # avoid unexpected error exception from timegm
   my @t=gmtime();
   $sec= $t[0] if ($sec<0||$sec>59);
   $min= $t[1] if ($min<0||$min>59);
   $hour=$t[2] if ($hour<0||$hour>23);
   $d   =$t[3] if ($d<1||$d>31);
   $m   =$t[4] if ($m<0||$m>11);
   $y   =$t[5] if ($y<70||$y>137);      # invalid if outside 1970...2037
   if ($d>28) {
      my @days_in_month = qw(0 31 28 31 30 31 30 31 31 30 31 30 31);
      my $year=1900+$y;
      $days_in_month[2]++ if ( $year%4==0 && ($year%100!=0||$year%400==0) );
      $d=$days_in_month[$m+1] if ($d>$days_in_month[$m+1]);
   }                                                                                                                                                 
   return timegm($sec,$min,$hour, $d,$m,$y);
}

sub is_dst {
   my ($gmtime, $timeoffset)=@_;
   my ($month,$year)=(seconds2array($gmtime))[4,5];     # $month 0..11
   my $seconds=timeoffset2seconds($timeoffset);
   
   my ($gm, $lt, $dow);
   if ($seconds >= -9*3600 && $seconds <= -3*3600 ) {   # dst rule for us
      return 1 if ($month>3 && $month<9);
      if ($month==3) {                                                                                                                               
         $lt=array2seconds(0,0,2, 1,3,$year);   # localtime Apr/1 2:00
         $dow=(seconds2array($lt))[6];          # weekday of localtime Apr/1 2:00:01
         $gm=$lt+(7-$dow)*86400-$seconds;       # gmtime of localtime Apr/1st Sunday
         return 1 if ($gmtime>=$gm);
      } elsif ($month==9) {
         $lt=array2seconds(0,0,2, 30,9,$year);  # localtime Oct/30 2:00
         $dow=(seconds2array($lt))[6];          # weekday of localtime Oct/30
         $gm=$lt-$dow*86400-$seconds;           # gmtime of localtime Oct/last Sunday
         return 1 if ($gmtime<=$gm);
      }
   } elsif ($seconds >= 0 && $seconds <= 6*3600 ) {     # dst rule for europe
      return 1 if ($month>2 && $month<9);
      if ($month==2) {
         $gm=array2seconds(0,0,1, 31,2,$year);  # gmtime Mar/31 1:00
         $dow=(seconds2array($gm))[6];          # weekday of gmtime Mar/31
         $gm-=$dow*86400;                       # gmtime Mar/last Sunday
         return 1 if ($gmtime>=$gm);
      } elsif ($month==9) {
         $gm=array2seconds(0,0,1, 30,9,$year);  # gmtime Oct/30 1:00
         $dow=(seconds2array($gm))[6];          # weekday of gmtime Oct/30
         $gm-=$dow*86400;                       # gmtime Oct/last Sunday
         return 1 if ($gmtime<=$gm);
      }
   }
   return 0;
}

sub time_gm2local {
   my ($g2l, $timeoffset, $daylightsaving)=@_;
   if ($daylightsaving eq 'on' ||
       ($daylightsaving eq 'auto' && is_dst($g2l,$timeoffset)) ) {
      $g2l+=3600; # plus 1 hour if is_dst at this gmtime
   }
   $g2l+=timeoffset2seconds($timeoffset) if ($timeoffset);
   return $g2l;
}

sub time_local2gm {
   my ($l2g, $timeoffset, $daylightsaving)=@_;
   $l2g-=timeoffset2seconds($timeoffset);
   if ($daylightsaving eq 'on' ||
       ($daylightsaving eq 'auto' && is_dst($l2g,$timeoffset)) ) {
      $l2g-=3600; # minus 1 hour if is_dst at that gmtime
   }
   return $l2g;
}

sub gmtime2dateserial {
   # time() is used if $_[0] undefined
   my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst)=seconds2array($_[0]||time());
   return(sprintf("%4d%02d%02d%02d%02d%02d", $year+1900, $mon+1, $mday, $hour, $min, $sec));
}

sub dateserial2gmtime {
   $_[0]=~/(\d\d\d\d)(\d\d)(\d\d)(\d\d)?(\d\d)?(\d\d)?/;
   my ($year, $mon, $mday, $hour, $min, $sec)=($1, $2, $3, $4, $5, $6);
   return array2seconds($sec,$min,$hour, $mday,$mon-1,$year-1900);
}

sub datefield2dateserial {      # return dateserial of GMT
   my $datefield=$_[0];
   my ($sec,$min,$hour, $mday,$mon,$year, $timeoffset,$timezone, $ampm);
   
   $sec = $min = $hour = $mday = $mon = $year = $timeoffset = $timezone = $ampm = "";
   
   $datefield=~s/GMT//;
   foreach my $s (split(/[\s,]+/, $datefield)) {
      if ($s=~/^\d\d?$/) {
         if ($s<=31 && $mday eq "") {
            $mday=$s;
         } else {
            $year=$s+1900;
            $year+=100 if ($year<1970);
	    $year-=1900;
         }
      } elsif ($s=~/^[A-Z][a-z][a-z]/i ) {
         for my $i (0..11) {
            if ($s=~/^$month_en[$i]/i) {
               $mon=$i+1; last;
            }
         }
      } elsif ($s=~/^\d\d\d\d$/) {
         $year=$s-1900;
      } elsif ($s=~/^(\d+):(\d+):?(\d+)?$/) {
         $hour=$1; $min=$2; $sec=$3?$3:0;
      } elsif ($s=~/^\(?([A-Z]{3,4}\d?)\)?$/) {
         $timezone=$1;
      } elsif ($s=~/^([\+\-]\d\d:?\d\d)$/) {
         $timeoffset=$1;
         $timeoffset=~s/://;
      } elsif ($s=~/^pm$/i) {
         $ampm='pm';
      }
   }
   $hour+=12 if ($hour && $hour<12 && $ampm eq 'pm');
   $timeoffset=$tzoffset{$timezone} if ($timeoffset eq "");
   if(!defined $timeoffset) { $timeoffset = 0; }
   unless($mon) { $mon = 0; }

   # NOTICE! The date field in msg header is generated by other machine
   #         Both datetime and the timezone str in date field include the dst shift,
   #         so we don't do daylightsaving here
   my $gm=time_local2gm(array2seconds($sec,$min,$hour, $mday,$mon-1,$year), $timeoffset, 0);
   return(gmtime2dateserial($gm));
}

sub dateserial2str {
   my ($dateserial, $timeoffset, $daylightsaving, $format, $hourformat)=@_;
   
   my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst)=
      seconds2array(time_gm2local(dateserial2gmtime($dateserial), $timeoffset, $daylightsaving));
   $year+=1900; $mon++;
   
   my $str;
   if ( $format eq "mm/dd/yyyy") {
      $str=sprintf("%02d/%02d/%04d", $mon, $mday, $year);
   } elsif ( $format eq "dd/mm/yyyy") {
      $str=sprintf("%02d/%02d/%04d", $mday, $mon, $year);
   } elsif ( $format eq "yyyy/mm/dd") {
      $str=sprintf("%04d/%02d/%02d", $year, $mon, $mday);
   
   } elsif ( $format eq "mm-dd-yyyy") {
      $str=sprintf("%02d-%02d-%04d", $mon, $mday, $year);
   } elsif ( $format eq "dd-mm-yyyy") {
      $str=sprintf("%02d-%02d-%04d", $mday, $mon, $year);
   } elsif ( $format eq "yyyy-mm-dd") {
      $str=sprintf("%04d-%02d-%02d", $year, $mon, $mday);
   
   } elsif ( $format eq "mm.dd.yyyy") {
      $str=sprintf("%02d.%02d.%04d", $mon, $mday, $year);
   } elsif ( $format eq "dd.mm.yyyy") {
      $str=sprintf("%02d.%02d.%04d", $mday, $mon, $year);                                                                                              
   } elsif ( $format eq "yyyy.mm.dd") {                                                                                                                
      $str=sprintf("%04d.%02d.%02d", $year, $mon, $mday);

   } else {
      $str=sprintf("%02d/%02d/%04d", $mon, $mday, $year);
   }

   if ( $hourformat eq "12") {
      my ($h, $ampm)=hour24to12($hour);
      $str.=sprintf(" %02d:%02d:%02d $ampm", $h, $min, $sec);
   } else {
      $str.=sprintf(" %02d:%02d:%02d", $hour, $min, $sec);                                                                                           
   }
   return($str);
}

sub hour24to12 {
   my $hour=$_[0];
   my $ampm="am";
   
   $hour =~ s/^0(.+)/$1/;
   if ($hour==24||$hour==0) {
      $hour = 12;
   } elsif ($hour > 12) {
      $hour = $hour - 12;
      $ampm = "pm";
   } elsif ($hour == 12) {
      $ampm = "pm";
   }
   return($hour, $ampm);
}

1;
